# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.backend.burp1
    :platform: Unix
    :synopsis: Burp-UI burp1 backend module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import re
import os
import socket
import time
import datetime
import json
import shutil
import subprocess
import tempfile

from .interface import BUIbackend
from ..parser.burp1 import Parser
from ...utils import human_readable as _hr, BUIcompress, utc_to_local
from ...security import sanitize_string
from ...exceptions import BUIserverException
from ..._compat import unquote, to_unicode, to_bytes

from shlex import quote


class Burp(BUIbackend):
    """The :class:`burpui.misc.backend.burp1.Burp` class provides a consistent
    backend for ``burp-1`` servers.

    It implements the :class:`burpui.misc.backend.interface.BUIbackend` class
    in order to have consistent data whatever backend is used.

    :param server: ``Burp-UI`` server instance in order to access logger
                   and/or some global settings
    :type server: :class:`burpui.engines.server.BUIServer`

    :param conf: Configuration file to use
    :type conf: str

    :param dummy: Does not instanciate the object (used for development
                  purpose)
    :type dummy: boolean
    """

    # backend version
    _vers = 1

    states = {
        "i": "idle",
        "r": "running",
        "c": "client crashed",
        "C": "server crashed",
        "1": "scanning",
        "2": "backup",
        "3": "merging",
        "4": "shuffling",
        "7": "listing",
        "8": "restoring",
        "9": "verifying",
        "0": "deleting",
    }

    counters = [
        "phase",
        "total",
        "files",
        "files_encrypted",
        "meta_data",
        "meta_data_encrypted",
        "directories",
        "soft_links",
        "hard_links",
        "special_files",
        "vss_headers",
        "vss_headers_encrypted",
        "vss_footers",
        "vss_footers_encrypted",
        "grand_total",
        "warning",
        "estimated_bytes",
        "bytes",
        "bytes_in",
        "bytes_out",
        "start",
        "path",
    ]

    def __init__(self, server=None, conf=None, dummy=False):
        """The :class:`burpui.misc.backend.burp1.Burp` class provides a consistent
        backend for ``burp-1`` servers.

        It implements the :class:`burpui.misc.backend.interface.BUIbackend` class
        in order to have consistent data whatever backend is used.

        :param server: ``Burp-UI`` server instance in order to access logger
                       and/or some global settings
        :type server: :class:`burpui.engines.server.BUIServer`

        :param conf: Configuration to use
        :type conf: :class:`burpui.config.BUIConfig`

        :param dummy: Does not instantiate the object (used for development
                      purpose)
        :type dummy: boolean
        """
        if dummy:
            return
        self.client_version = ""
        self.server_version = ""

        BUIbackend.__init__(self, server, conf)

        self.parser = Parser(self)

        self.family = Burp._get_inet_family(self.host)
        self._test_burp_server_address(self.host)

        try:
            cmd = [self.burpbin, "-v"]
            self.client_version = (
                subprocess.check_output(cmd).rstrip().replace("burp-", "")
            )
        except:
            pass

        try:  # pragma: no cover
            cmd = [self.burpbin, "-a", "l"]
            if self.burpconfcli:
                cmd += ["-c", self.burpconfcli]
            for line in subprocess.check_output(cmd).split("\n"):
                result = re.search(r"^.*Server version:\s+(\d+\.\d+\.\d+)", line)
                if result:
                    self.server_version = result.group(1)
                    break
        except:
            pass

        self.logger.info(f"burp port: {self.port}")
        self.logger.info(f"burp host: {self.host}")
        self.logger.info(f"burp binary: {self.burpbin}")
        self.logger.info(f"strip binary: {self.stripbin}")
        self.logger.info(f"burp conf cli: {self.burpconfcli}")
        self.logger.info(f"burp conf srv: {self.burpconfsrv}")
        self.logger.info(f"command timeout: {self.timeout}")
        self.logger.info(f"tmpdir: {self.tmpdir}")
        self.logger.info(f"zip64: {self.zip64}")
        self.logger.info(f"includes: {self.includes}")
        self.logger.info(f"enforce: {self.enforce}")
        self.logger.info(f"revoke: {self.revoke}")
        self.logger.info(f"client version: {self.client_version}")
        self.logger.info(f"server version: {self.server_version}")
        try:
            # make the connection
            self.status()
        except BUIserverException:
            pass

    @staticmethod
    def _get_inet_family(addr):
        """The :func:`burpui.misc.backend.burp1.Burp._get_inet_family` function
        determines the inet family of a given address.

        :param addr: Address to look at
        :type addr: str

        :returns: Inet family of the given address: :const:`socket.AF_INET` of
                  :const:`socket.AF_INET6`
        """
        if addr == "127.0.0.1":
            return socket.AF_INET
        else:
            return socket.AF_INET6

    def _test_burp_server_address(self, addr, retry=False):
        """The :func:`burpui.misc.backend.burp1.Burp._test_burp_server_address`
        function determines if the given address is reachable or not.

        :param addr: Address to look at
        :type addr: str

        :param retry: Flag to stop trying because this function is recursive
        :type retry: bool

        :returns: True or False whether we could find a valid address or not
        """
        family = Burp._get_inet_family(addr)
        try:
            sock = socket.socket(family, socket.SOCK_STREAM)
            sock.connect((addr, self.port))
            sock.close()
            return True
        except socket.error:
            self.logger.warning("Cannot contact burp server at %s:%s", addr, self.port)
            if not retry:
                new_addr = ""
                if self.host == "127.0.0.1":
                    new_addr = "::1"
                else:
                    new_addr = "127.0.0.1"
                self.logger.info("Trying %s:%s instead", new_addr, self.port)
                if self._test_burp_server_address(new_addr, True):
                    self.logger.info(
                        "%s:%s is reachable, switching to it for this runtime",
                        new_addr,
                        self.port,
                    )
                    self.host = new_addr
                    self.family = Burp._get_inet_family(new_addr)
                    return True
                self.logger.error("Cannot guess burp server address")
        return False

    def statistics(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.statistics`"""
        return {
            "alive": self._test_burp_server_address(self.host),
            "client_version": self.client_version,
            "server_version": self.server_version,
        }

    def status(self, query="\n", timeout=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.status`"""
        result = []
        try:
            query = query.rstrip()
            self.logger.info("query: '{}'".format(query))
            # FIXME: cannot sanitize string due to unicode :/
            # NOTE: converting it to bytes should minimize the risks though
            # query = sanitize_string(query.rstrip())
            qry = to_bytes("{}\n".format(query))
            sock = socket.socket(self.family, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
            sock.send(qry)
            sock.shutdown(socket.SHUT_WR)
            fileobj = sock.makefile()
            sock.close()
            for line in fileobj.readlines():
                line = line.rstrip("\n")
                if not line:
                    continue
                try:
                    line = to_unicode(line)
                except UnicodeDecodeError:  # pragma: no cover
                    pass
                result.append(line)
            fileobj.close()
            self.logger.debug("=> {}".format(result))
            return result
        except socket.error:
            self.logger.error(
                "Cannot contact burp server at {0}:{1}".format(self.host, self.port)
            )
            raise BUIserverException(
                "Cannot contact burp server at {0}:{1}".format(self.host, self.port)
            )

    def _get_all_backup_logs(self, client, forward=False, deep=False):
        ret = []
        backups = self.get_client(client)
        queue = []
        for back in backups:
            queue.append(self._get_backup_logs(back["number"], client, forward, deep))

        ret = sorted(queue, key=lambda x: x["number"])
        return ret

    def _get_backup_logs(self, number, client, forward=False, deep=False):
        if not client or not number:
            return {}

        filemap = self.status("c:{0}:b:{1}\n".format(client, number))
        ret = {}
        for line in filemap:
            if line == "backup_stats":
                ret = self._parse_backup_stats(number, client, forward)
                break
        else:
            cli = None
            if forward:
                cli = client

            filemap = self.status("c:{0}:b:{1}:f:log.gz\n".format(client, number))
            ret = self._parse_backup_log(filemap, number, cli)

        ret["encrypted"] = False
        if "files_enc" in ret and ret["files_enc"]["total"] > 0:
            ret["encrypted"] = True
        return ret

    def get_backup_logs(self, number, client, forward=False, deep=False, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_backup_logs`"""
        if not client or not number:
            return {} if number and number != -1 else []

        if number == -1:
            return self._get_all_backup_logs(client, forward, deep)
        return self._get_backup_logs(number, client, forward, deep)

    def _parse_backup_stats(
        self, number, client, forward=False, stats=None, agent=None
    ):
        """The :func:`burpui.misc.backend.burp1.Burp._parse_backup_stats`
        function is used to parse the burp logs.

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :param forward: Is the client name needed in later process
        :type forward: bool

        :param agent: What server to ask (only in multi-agent mode)
        :type agent: str

        :returns: Dict containing the backup log
        """
        backup = {"os": "unknown", "number": int(number)}
        if forward:
            backup["name"] = client
        keys = {
            "time_start": "start",
            "time_end": "end",
            "time_taken": "duration",
            "bytes_in_backup": "totsize",
            "bytes_received": "received",
            "files": ["files", "new"],
            "files_changed": ["files", "changed"],
            "files_same": ["files", "unchanged"],
            "files_deleted": ["files", "deleted"],
            "files_scanned": ["files", "scanned"],
            "files_total": ["files", "total"],
            "files_encrypted": ["files_enc", "new"],
            "files_encrypted_changed": ["files_enc", "changed"],
            "files_encrypted_same": ["files_enc", "unchanged"],
            "files_encrypted_deleted": ["files_enc", "deleted"],
            "files_encrypted_scanned": ["files_enc", "scanned"],
            "files_encrypted_total": ["files_enc", "total"],
            "directories": ["dir", "new"],
            "directories_changed": ["dir", "changed"],
            "directories_same": ["dir", "unchanged"],
            "directories_deleted": ["dir", "deleted"],
            "directories_scanned": ["dir", "scanned"],
            "directories_total": ["dir", "total"],
            "soft_links": ["softlink", "new"],
            "soft_links_changed": ["softlink", "changed"],
            "soft_links_same": ["softlink", "unchanged"],
            "soft_links_deleted": ["softlink", "deleted"],
            "soft_links_scanned": ["softlink", "scanned"],
            "soft_links_total": ["softlink", "total"],
            "hard_links": ["hardlink", "new"],
            "hard_links_changed": ["hardlink", "changed"],
            "hard_links_same": ["hardlink", "unchanged"],
            "hard_links_deleted": ["hardlink", "deleted"],
            "hard_links_scanned": ["hardlink", "scanned"],
            "hard_links_total": ["hardlink", "total"],
            "meta_data": ["meta", "new"],
            "meta_data_changed": ["meta", "changed"],
            "meta_data_same": ["meta", "unchanged"],
            "meta_data_deleted": ["meta", "deleted"],
            "meta_data_scanned": ["meta", "scanned"],
            "meta_data_total": ["meta", "total"],
            "meta_data_encrypted": ["meta_enc", "new"],
            "meta_data_encrypted_changed": ["meta_enc", "changed"],
            "meta_data_encrypted_same": ["meta_enc", "unchanged"],
            "meta_data_encrypted_deleted": ["meta_enc", "deleted"],
            "meta_data_encrypted_scanned": ["meta_enc", "scanned"],
            "meta_data_encrypted_total": ["meta_enc", "total"],
            "special_files": ["special", "new"],
            "special_files_changed": ["special", "changed"],
            "special_files_same": ["special", "unchanged"],
            "special_files_deleted": ["special", "deleted"],
            "special_files_scanned": ["special", "scanned"],
            "special_files_total": ["special", "total"],
            "efs_files": ["efs", "new"],
            "efs_files_changed": ["efs", "changed"],
            "efs_files_same": ["efs", "unchanged"],
            "efs_files_deleted": ["efs", "deleted"],
            "efs_files_scanned": ["efs", "scanned"],
            "efs_files_total": ["efs", "total"],
            "vss_headers": ["vssheader", "new"],
            "vss_headers_changed": ["vssheader", "changed"],
            "vss_headers_same": ["vssheader", "unchanged"],
            "vss_headers_deleted": ["vssheader", "deleted"],
            "vss_headers_scanned": ["vssheader", "scanned"],
            "vss_headers_total": ["vssheader", "total"],
            "vss_headers_encrypted": ["vssheader_enc", "new"],
            "vss_headers_encrypted_changed": ["vssheader_enc", "changed"],
            "vss_headers_encrypted_same": ["vssheader_enc", "unchanged"],
            "vss_headers_encrypted_deleted": ["vssheader_enc", "deleted"],
            "vss_headers_encrypted_scanned": ["vssheader_enc", "scanned"],
            "vss_headers_encrypted_total": ["vssheader_enc", "total"],
            "vss_footers": ["vssfooter", "new"],
            "vss_footers_changed": ["vssfooter", "changed"],
            "vss_footers_same": ["vssfooter", "unchanged"],
            "vss_footers_deleted": ["vssfooter", "deleted"],
            "vss_footers_scanned": ["vssfooter", "scanned"],
            "vss_footers_total": ["vssfooter", "total"],
            "vss_footers_encrypted": ["vssfooter_enc", "new"],
            "vss_footers_encrypted_changed": ["vssfooter_enc", "changed"],
            "vss_footers_encrypted_same": ["vssfooter_enc", "unchanged"],
            "vss_footers_encrypted_deleted": ["vssfooter_enc", "deleted"],
            "vss_footers_encrypted_scanned": ["vssfooter_enc", "scanned"],
            "vss_footers_encrypted_total": ["vssfooter_enc", "total"],
            "total": ["total", "new"],
            "total_changed": ["total", "changed"],
            "total_same": ["total", "unchanged"],
            "total_deleted": ["total", "deleted"],
            "total_scanned": ["total", "scanned"],
            "total_total": ["total", "total"],
        }
        if not stats:
            filemap = self.status(
                "c:{0}:b:{1}:f:backup_stats\n".format(client, number), agent=agent
            )
        else:
            filemap = stats
        for line in filemap:
            if line == "-list begin-" or line == "-list end-":
                continue
            (key, val) = line.split(":")
            if backup["os"] == "unknown" and key == "client_is_windows":
                if val == "1":
                    backup["os"] = "Windows"
                else:
                    backup["os"] = "Unix/Linux"
                continue
            if key not in keys:
                continue
            ckey = keys[key]
            if isinstance(ckey, list):
                if ckey[0] not in backup:
                    backup[ckey[0]] = {}
                backup[ckey[0]][ckey[1]] = int(val)
            else:
                backup[ckey] = int(val)

        # Needed for graphs
        if "received" not in backup:
            backup["received"] = 1

        return backup

    def _parse_backup_log(self, filemap, number, client=None, agent=None):
        """The :func:`burpui.misc.backend.burp1.Burp._parse_backup_log` function
        is used to parse the log.gz of a given backup and returns a dict
        containing different stats used to render the charts in the reporting
        view.

        :param filemap: List representing the content of the log file
        :type filemap: list

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :param agent: What server to ask (only in multi-agent mode)
        :type agent: str

        :returns: Dict containing the backup log
        """
        lookup_easy = {
            "start": r"^Start time: (.+)$",
            "end": r"^\s*End time: (.+)$",
            "duration": r"^Time taken: (.+)$",
            "totsize": r"^\s*Bytes in backup:\s+(\d+)",
            "received": r"^\s*Bytes received:\s+(\d+)",
        }
        lookup_complex = {
            "files": r"^\s*Files:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "files_enc": r"^\s*Files \(encrypted\):?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "dir": r"^\s*Directories:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "softlink": r"^\s*Soft links:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "hardlink": r"^\s*Hard links:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "meta": r"^\s*Meta data:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "meta_enc": r"^\s*Meta data\(enc\):?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "special": r"^\s*Special files:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "efs": r"^\s*EFS files:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "vssheader": r"^\s*VSS headers:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "vssheader_enc": r"^\s*VSS headers \(enc\):?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "vssfooter": r"^\s*VSS footers:?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "vssfooter_enc": r"^\s*VSS footers \(enc\):?\s+([\d\s]+)\s+\|\s+(\d+)$",
            "total": r"^\s*Grand total:?\s+([\d\s]+)\s+\|\s+(\d+)$",
        }
        _ = agent  # noqa
        backup = {"os": "Unix/Linux", "number": int(number)}
        if client is not None:
            backup["name"] = client
        useful = False
        for line in filemap:
            if re.match(
                r"^\d{4}-\d{2}-\d{2} (\d{2}:){3} \w+\[\d+\] Client is Windows$", line
            ):
                backup["os"] = "Windows"
            elif not useful and not re.match(r"^-+$", line):
                continue
            elif useful and re.match(r"^-+$", line):
                useful = False
                continue
            elif re.match(r"^-+$", line):
                useful = True
                continue

            found = False
            # this method is not optimal, but it is easy to read and to maintain
            for (key, regex) in lookup_easy.items():
                reg = re.search(regex, line)
                if reg:
                    found = True
                    if key in ["start", "end"]:
                        backup[key] = int(
                            time.mktime(
                                datetime.datetime.strptime(
                                    reg.group(1), "%Y-%m-%d %H:%M:%S"
                                ).timetuple()
                            )
                        )
                    elif key == "duration":
                        tmp = reg.group(1).split(":")
                        tmp.reverse()
                        fields = [0] * 4
                        for (i, val) in enumerate(tmp):
                            fields[i] = int(val)
                        seconds = 0
                        seconds += fields[0]
                        seconds += fields[1] * 60
                        seconds += fields[2] * (60 * 60)
                        seconds += fields[3] * (60 * 60 * 24)
                        backup[key] = seconds
                    else:
                        backup[key] = int(reg.group(1))
                    # break the loop as soon as we find a match
                    break

            # if found is True, we already parsed the line so we can jump to the next one
            if found:
                continue

            for (key, regex) in lookup_complex.items():
                reg = re.search(regex, line)
                if reg:
                    # self.logger.debug("match[1]: '{0}'".format(reg.group(1)))
                    spl = re.split(r"\s+", reg.group(1))
                    if len(spl) < 5:
                        return {}
                    backup[key] = {
                        "new": int(spl[0]),
                        "changed": int(spl[1]),
                        "unchanged": int(spl[2]),
                        "deleted": int(spl[3]),
                        "total": int(spl[4]),
                        "scanned": int(reg.group(2)),
                    }
                    break
        return backup

    def get_clients_report(self, clients, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_clients_report`"""
        data = []
        for cli in clients:
            if not cli:
                continue
            client = self.get_client(cli["name"])
            if not client or not client[-1]:
                continue
            stats = self.get_backup_logs(client[-1]["number"], cli["name"])
            data.append((cli, client, stats))
        return self._do_get_clients_report(data)

    def _do_get_clients_report(self, data):
        ret = {}
        clients = []
        bkp = []
        for client, backups, stats in data:
            os = stats["os"] if "os" in stats else "unknown"
            totsize = stats["totsize"] if "totsize" in stats else 0
            total = (
                stats["total"]["total"]
                if "total" in stats and "total" in stats["total"]
                else 0
            )
            clients.append(
                {
                    "name": client["name"],
                    "stats": {"os": os, "totsize": totsize, "total": total},
                }
            )
            bkp.append({"name": client["name"], "number": len(backups)})
        ret = {"clients": clients, "backups": bkp}
        return ret

    def get_counters(
        self, name=None, agent=None
    ):  # pragma: no cover (hard to test, requires a running backup)
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_counters`"""
        res = {}
        filemap = self.status("c:{0}\n".format(name))
        if not filemap:
            return res
        for line in filemap:
            # self.logger.debug('line: {0}'.format(line))
            reg = re.search(r"^{0}\s+(\d)\s+(\S)\s+(.+)$".format(name), line)
            if reg and reg.group(2) == "r" and int(reg.group(1)) == 2:
                count = 0
                for val in reg.group(3).split("\t"):
                    # self.logger.debug('{0}: {1}'.format(self.counters[c], v))
                    if val and count > 0 and count < 15:
                        try:
                            vals = list(map(int, val.split("/")))
                            if vals[0] > 0 or vals[1] > 0 or vals[2] or vals[3] > 0:
                                res[self.counters[count]] = vals
                        except (ValueError, IndexError):
                            count += 1
                            continue
                    elif val:
                        if self.counters[count] == "path":
                            res[self.counters[count]] = val
                        else:
                            try:
                                res[self.counters[count]] = int(val)
                            except ValueError:
                                count += 1
                                continue
                    count += 1

        # the client was not running
        if not res:
            return res

        if "bytes" not in res:
            res["bytes"] = 0
        if set(["start", "estimated_bytes", "bytes_in"]) <= set(res.keys()):
            try:
                diff = time.time() - int(res["start"])
                byteswant = int(res["estimated_bytes"])
                bytesgot = int(res["bytes_in"])
                bytespersec = bytesgot / diff
                bytesleft = byteswant - bytesgot
                res["speed"] = bytespersec
                if bytespersec > 0:
                    timeleft = int(bytesleft / bytespersec)
                    res["timeleft"] = timeleft
                else:
                    res["timeleft"] = -1
            except:
                res["timeleft"] = -1
        try:
            res["percent"] = round(
                float(res["bytes"]) / float(res["estimated_bytes"]) * 100
            )
        except ZeroDivisionError:
            # You know... division by 0
            res["percent"] = 0
        return res

    def is_backup_running(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.is_backup_running`"""
        if not name:
            return False
        try:
            filemap = self.status("c:{0}\n".format(name))
        except BUIserverException:
            return False
        for line in filemap:
            reg = re.search(r"^{0}\s+\d\s+(\w)".format(name), line)
            if reg and reg.group(1) not in ["i", "c", "C"]:
                return True
        return False

    def is_one_backup_running(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.is_one_backup_running`"""
        res = []
        try:
            clients = self.get_all_clients(last_attempt=False)
        except BUIserverException:
            return res
        for client in clients:
            if self.is_backup_running(client["name"]):
                res.append(client["name"])
        return res

    def get_all_clients(self, agent=None, last_attempt=True):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_all_clients`"""
        res = []
        filemap = self.status()
        for line in filemap:
            regex = re.compile(r"\s*(\S+)\s+\d\s+(\S)\s+(.+)")
            match = regex.search(line)
            cli = {}
            cli["name"] = match.group(1)
            cli["state"] = self.states[match.group(2)]
            infos = match.group(3)
            if cli["state"] in ["running"]:
                regex = re.compile(r"\s*(\S+)")
                reg = regex.search(infos)
                phase = reg.group(0)
                if phase and phase in self.states:
                    cli["phase"] = self.states[phase]
                else:
                    cli["phase"] = "unknown"
                cli["last"] = "now"
                counters = self.get_counters(cli["name"])
                if "percent" in counters:
                    cli["percent"] = counters["percent"]
                else:
                    cli["percent"] = 0
            elif infos == "0":
                cli["last"] = "never"
            elif re.match(r"^\d+\s\d+\s\d+$", infos):
                spl = infos.split()
                cli["last"] = int(spl[2])
            else:
                spl = infos.split("\t")
                cli["last"] = int((spl[-1].split())[-1])
            cli["last"] = utc_to_local(cli["last"])
            cli["last_attempt"] = cli["last"]
            res.append(cli)
        return res

    def get_client_status(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_status`"""
        cli = {}
        filemap = self.status("c:{0}\n".format(name))
        for line in filemap:
            if not re.match("^{0}\t".format(name), line):
                continue
            regex = re.compile(r"\s*(\S+)\s+\d\s+(\S)\s+(.+)")
            match = regex.search(line)
            cli["state"] = self.states[match.group(2)]
            infos = match.group(3)
            if cli["state"] in ["running"]:
                regex = re.compile(r"\s*(\S+)")
                reg = regex.search(infos)
                phase = reg.group(0)
                if phase and phase in self.states:
                    cli["phase"] = self.states[phase]
                else:
                    cli["phase"] = "unknown"
                cli["last"] = "now"
                counters = self.get_counters(name)
                if "percent" in counters:
                    cli["percent"] = counters["percent"]
                else:
                    cli["percent"] = 0
            elif infos == "0":
                cli["last"] = "never"
            elif re.match(r"^\d+\s\d+\s\d+$", infos):
                spl = infos.split()
                cli["last"] = int(spl[2])
            else:
                spl = infos.split("\t")
                cli["last"] = int((spl[-1].split())[-1])
            cli["last"] = utc_to_local(cli["last"])
            break
        return cli

    def get_client(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client`"""
        return self.get_client_filtered(name)

    def get_client_filtered(
        self, name=None, limit=-1, page=None, start=None, end=None, agent=None
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_filtered`"""
        res = []
        if not name:
            return res
        cli = name
        filemap = self.status("c:{0}\n".format(cli))
        for line in filemap:
            if not re.match("^{0}\t".format(cli), line):
                continue
            # self.logger.debug("line: '{0}'".format(line))
            regex = re.compile(r"\s*(\S+)\s+\d\s+(\S)\s+(.+)")
            match = regex.search(line)
            if match.group(3) == "0" or match.group(2) not in ["i", "c", "C"]:
                continue
            backups = match.group(3).split("\t")
            for cpt, backup in enumerate(backups):
                # skip the first elements if we are in a page
                if page and page > 1 and limit > 0:
                    if cpt < (page - 1) * limit:
                        continue
                bkp = {}
                spl = backup.split()
                backup_date = int(spl[2])
                bkp["number"] = spl[0]
                bkp["deletable"] = spl[1] == "1"
                bkp["date"] = utc_to_local(backup_date)
                # skip backups before "start"
                if start and backup_date < start:
                    continue
                # skip backups after "end"
                if end and backup_date > end:
                    continue
                log = self.get_backup_logs(spl[0], name)
                bkp["encrypted"] = log["encrypted"]
                bkp["received"] = log["received"]
                bkp["size"] = log["totsize"]
                bkp["end"] = utc_to_local(log["end"])
                res.append(bkp)
                # stop after "limit" elements
                if page and page > 1 and limit > 0:
                    if cpt >= page * limit:
                        break
                if limit > 0 and cpt >= limit:
                    break
        # Here we need to reverse the array so the backups are sorted by num ASC
        res.reverse()
        return res

    def is_backup_deletable(self, name=None, backup=None, agent=None):
        """Check if a given backup is deletable"""
        # This feature won't be available in the burp 1 backend so we always
        # return False
        return False

    def get_tree(self, name=None, backup=None, root=None, level=-1, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_tree`"""
        res = []
        if not name or not backup:
            return res
        if not root:
            top = ""
        else:
            top = to_unicode(root)

        filemap = self.status("c:{0}:b:{1}:p:{2}\n".format(name, backup, top))
        useful = False
        for line in filemap:
            if not useful and re.match(r"^-list begin-$", line):
                useful = True
                continue
            if useful and re.match(r"^-list end-$", line):
                useful = False
                continue
            if useful:
                tree = {}
                match = re.search(r"^(.{10})\s", line)
                if match:
                    if re.match(r"^(d|l)", match.group(1)):
                        tree["type"] = "d"
                        tree["folder"] = True
                    else:
                        tree["type"] = "f"
                        tree["folder"] = False
                    spl = re.split(r"\s+", line, 7)
                    tree["mode"] = spl[0]
                    tree["inodes"] = spl[1]
                    tree["uid"] = spl[2]
                    tree["gid"] = spl[3]
                    tree["size"] = "{0:.1eM}".format(_hr(spl[4]))
                    tree["date"] = "{0} {1}".format(spl[5], spl[6])
                    tree["name"] = spl[7]
                    tree["parent"] = top
                    tree["fullname"] = os.path.join(top, spl[7])
                    tree["level"] = level
                    tree["children"] = []
                    res.append(tree)
        return res

    def is_server_restore(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.is_server_restore`"""
        return self.parser.read_restore(client)

    def cancel_server_restore(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.cancel_server_restore`"""
        return self.parser.cancel_restore(client)

    def server_restore(
        self,
        client=None,
        backup=None,
        files=None,
        strip=None,
        force=None,
        prefix=None,
        restoreto=None,
        agent=None,
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.server_restore`"""
        if not client or not backup or not files:
            raise BUIserverException("At least one argument is missing")

        # TODO: maybe someday we'd like to block server initiated restore
        # if not self.parser.client_conf.get('server_can_restore'):
        #     raise BUIserverException('Server initiated restoration is disabled')

        return self.parser.server_initiated_restoration(
            client, backup, files, strip, force, prefix, restoreto
        )

    def is_server_backup(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.is_server_backup`"""
        try:
            return self.parser.read_backup(client)
        except BUIserverException:
            return None

    def cancel_server_backup(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.cancel_server_backup`"""
        try:
            return self.parser.cancel_backup(client)
        except BUIserverException:
            return None

    def server_backup(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.server_backup`"""
        try:
            return self.parser.server_initiated_backup(client)
        except BUIserverException:
            return None

    def delete_backup(self, name=None, backup=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.delete_backup`"""
        if self._vers == 1:
            return "Sorry, this feature is not available"
        if not name or not backup:
            return "At least one argument is missing"
        if not self.burpbin:
            return "Missing 'burp' binary"
        if not self.is_backup_deletable(name, backup):
            return "Sorry, this backup is not deletable"
        cmd = [
            self.burpbin,
            "-C",
            quote(name),
            "-a",
            "delete",
            "-b",
            quote(str(backup)),
            "-c",
            self.burpconfcli,
        ]
        self.logger.debug(cmd)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        out, _ = proc.communicate()
        status = proc.wait()
        self.logger.debug(out)
        self.logger.debug("command returned: %d", status)
        if status != 0:
            return "The command failed with status {}: {}".format(status, out)
        return None

    def restore_files(
        self,
        name=None,
        backup=None,
        files=None,
        strip=None,
        archive="zip",
        password=None,
        agent=None,
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.restore_files`"""
        if not name or not backup or not files:
            return None, "At least one argument is missing"
        if not self.stripbin:
            return None, "Missing 'strip' binary"
        if not self.burpbin:
            return None, "Missing 'burp' binary"
        if not self.tmpdir:
            return None, "Missing 'tmpdir'"
        flist = json.loads(files)
        if password:
            tmphandler, tmpfile = tempfile.mkstemp()
        if not os.path.exists(self.tmpdir):
            os.makedirs(self.tmpdir)
        tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        if "restore" not in flist:
            return None, "Wrong call"
        if os.path.isdir(tmpdir):
            shutil.rmtree(tmpdir)
        full_reg = ""

        def _escape(s):
            return re.sub(r"[(){}\[\].*?|^$\\+-]", r"\\\g<0>", s)

        for restore in flist["restore"]:
            reg = ""
            if restore["folder"] and restore["key"] != "/":
                reg += "^" + _escape(restore["key"]) + "/|"
            else:
                reg += "^" + _escape(restore["key"]) + "$|"
            full_reg += to_unicode(reg)

        cmd = [
            self.burpbin,
            "-C",
            quote(name),
            "-a",
            "r",
            "-b",
            quote(str(backup)),
            "-r",
            full_reg.rstrip("|").replace(r"\n", r"\\n"),
            "-d",
            tmpdir,
        ]
        if password:
            if not self.burpconfcli:
                return None, "No client configuration file specified"
            tmpdesc = os.fdopen(tmphandler, "wb+")
            with open(self.burpconfcli, "rb") as fileobj:
                shutil.copyfileobj(fileobj, tmpdesc)

            tmpdesc.write(
                to_bytes("encryption_password = {}\n".format(sanitize_string(password)))
            )
            tmpdesc.close()
            cmd.append("-c")
            cmd.append(tmpfile)
        elif self.burpconfcli:
            cmd.append("-c")
            cmd.append(self.burpconfcli)
        if strip and strip > 0:  # 0 is False, but we are sure now
            cmd.append("-s")
            cmd.append(str(strip))
        self.logger.debug(cmd)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        out, _ = proc.communicate()
        status = proc.wait()
        if password:
            os.remove(tmpfile)
        self.logger.debug(out)
        self.logger.debug("command returned: %d", status)
        # hack to handle client-side encrypted backups
        # this is now handled client-side, but we should never trust user input
        # so we need to handle it server-side too
        decode = to_unicode(out)
        if (
            "zstrm inflate error: -3" in decode
            and "transfer file returning: -1" in decode
        ):
            status = 1
            out = "encrypted"
        # a return code of 2 means there were some warnings during restoration
        # so we can assume the restoration was successful anyway
        if status not in [0, 2]:
            return None, out

        if not os.path.isdir(tmpdir):
            return None, "Nothing to restore"

        zip_dir = tmpdir.rstrip(os.sep)
        zip_file = zip_dir + ".zip"
        if os.path.isfile(zip_file):
            os.remove(zip_file)
        zip_len = len(zip_dir) + 1
        stripping = True
        test_strip = True
        with BUIcompress(zip_file, archive, self.zip64) as zfh:
            for dirname, _, files in os.walk(zip_dir):
                for filename in files:
                    path = os.path.join(dirname, filename)
                    # try to detect if the file contains vss headers
                    if test_strip:
                        test_strip = False
                        otp = None
                        try:
                            with open(os.devnull, "w") as devnul:
                                otp = subprocess.check_output(
                                    [self.stripbin, "-p", "-i", path], stderr=devnul
                                )
                        except subprocess.CalledProcessError as exc:
                            self.logger.debug(
                                "Stripping failed on '{}': {}".format(path, str(exc))
                            )
                        if not otp:
                            stripping = False

                    if stripping and os.path.isfile(path):
                        self.logger.debug("stripping file: %s", path)
                        shutil.move(path, path + ".tmp")
                        status = subprocess.call(
                            [self.stripbin, "-i", path + ".tmp", "-o", path]
                        )
                        if status != 0:
                            os.remove(path)
                            shutil.move(path + ".tmp", path)
                            stripping = False
                            self.logger.debug(
                                "Disable stripping since this file does not seem to embed VSS headers"
                            )
                        else:
                            os.remove(path + ".tmp")

                    entry = path[zip_len:]
                    zfh.append(path, entry)

        shutil.rmtree(tmpdir)
        return zip_file, None

    def read_conf_cli(self, client=None, conf=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.read_conf_cli`"""
        if not self.parser:
            return []
        return self.parser.read_client_conf(client, conf)

    def read_conf_srv(self, conf=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.read_conf_srv`"""
        if not self.parser:
            return []
        return self.parser.read_server_conf(conf)

    def store_conf_cli(
        self,
        data,
        client=None,
        conf=None,
        template=False,
        statictemplate=False,
        content="",
        agent=None,
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.store_conf_cli`"""
        if not self.parser:
            return []
        try:
            conf = unquote(conf)
        except:
            pass
        return self.parser.store_client_conf(
            data, client, conf, template, statictemplate, content
        )

    def store_conf_srv(self, data, conf=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.store_conf_srv`"""
        if not self.parser:
            return []
        try:
            conf = unquote(conf)
        except:
            pass
        return self.parser.store_conf(data, conf)

    def expand_path(self, path=None, source=None, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.expand_path`"""
        if not path:
            return []
        return self.parser.path_expander(path, source, client)

    def delete_client(
        self, client=None, keepconf=False, delcert=False, revoke=False, agent=None
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.delete_client`"""
        if not client:
            return [[2, "No client provided"]]
        return self.parser.remove_client(client, keepconf, delcert, revoke)

    def clients_list(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.clients_list`"""
        return self.parser.list_clients()

    def get_parser_attr(self, attr=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_parser_attr`"""
        if not attr or not self.parser:
            raise BUIserverException("Wrong call")
        return getattr(self.parser, attr, [])

    def revocation_enabled(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.revocation_enabled`
        """
        return self.revoke

    def get_client_version(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_version`"""
        return self.client_version

    def get_server_version(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_server_version`"""
        return self.server_version

    def get_client_labels(self, client=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_labels`"""
        # Not supported with Burp 1.x.x so we just return an empty list
        return []

    def get_attr(self, name, default=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_attr`"""
        try:
            return getattr(self, name, default)
        except AttributeError:
            return default

    def get_parser(self, assume_version=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_parser`"""
        if self.parser:
            if assume_version:
                self.parser._assume_server_version = assume_version
            return self.parser
        raise BUIserverException("Missing parser")

    def get_file(self, path, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_file`"""
        # only used by the multi backend. We do nothing
        return path

    def del_file(self, path, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.del_file`"""
        # only used by the multi backend. We do nothing
        return path

    def version(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.version`"""
        return self._vers



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.desc
    :platform: Unix
    :synopsis: Burp-UI desc module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os

__title__ = "burp-ui"
__author__ = "Benjamin SANS (Ziirish)"
__author_email__ = "hi+burpui@ziirish.me"
__url__ = "https://git.ziirish.me/ziirish/burp-ui"
__doc__ = "https://burp-ui.readthedocs.io/en/latest/"
__description__ = (
    "Burp-UI is a web-ui for burp backup written in python with "
    "Flask and jQuery/Bootstrap"
)
__license__ = "BSD 3-clause"
__version__ = (
    open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "VERSION"))
    .read()
    .rstrip()
)
try:  # pragma: no cover
    __release__ = (
        open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "RELEASE"))
        .read()
        .rstrip()
    )
except:  # pragma: no cover
    __release__ = "unknown"



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.cache
    :platform: Unix
    :synopsis: Burp-UI external cache module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask_caching import Cache


cache = Cache(
    config={
        "CACHE_TYPE": "simple",
        "CACHE_THRESHOLD": 50,
        "CACHE_DEFAULT_TIMEOUT": 7200,
    }
)



# -*- coding: utf8 -*-
import re
import logging

from .interface import BUIaudit, BUIauditLogger as BUIauditLoggerInterface
from ...tools.logging import logger as parent_logger


class BUIauditLoader(BUIaudit):
    section = name = "BASIC:AUDIT"

    logfile = None
    max_bytes = None
    rotate = None

    def __init__(self, app):
        """
        :param app: Instance of the app we are running in
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = app
        self.conf = self.app.conf

        self.level = default = logging.getLevelName(self.app.logger.getEffectiveLevel())

        if self.section in self.conf.options:
            self.priority = self.conf.safe_get(
                "priority", "integer", self.section, defaults=self.section
            )
            self.level = self.conf.safe_get(
                "level", section=self.section, defaults=default
            )
            self.logfile = self.conf.safe_get("logfile", section=self.section)
            self.max_bytes = self.conf.safe_get(
                "max_bytes",
                "force_string",
                section=self.section,
                defaults="30 * 1024 * 1024",
            )
            self.rotate = self.conf.safe_get(
                "rotate", "integer", section=self.section, defaults=5
            )

        if self.max_bytes and re.match(r"(\d+\s*[+-/*]?\s*)+$", self.max_bytes):
            self.max_bytes = eval(self.max_bytes)
        else:
            self.max_bytes = 0

        if self.level != default:
            self.level = logging.getLevelName(f"{self.level}".upper())
            if not isinstance(self.level, int):
                self.level = default

        self._logger = BUIauditLogger(self)


class BUIauditLogger(BUIauditLoggerInterface):
    _logger = parent_logger.getChild("audit")  # type: logging.Logger

    def __init__(self, loader):
        self.loader = loader
        self._level = self.loader.level
        LOG_FORMAT = "[%(asctime)s] AUDIT %(levelname)s in %(from)s: %(message)s"

        if self.loader.logfile and self.loader.logfile.lower() != "none":
            from logging.handlers import RotatingFileHandler

            handler = RotatingFileHandler(
                self.loader.logfile,
                maxBytes=self.loader.max_bytes,
                backupCount=self.loader.rotate,
            )
        else:
            from logging import StreamHandler

            handler = StreamHandler()

        handler.setLevel(self.level)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))

        self._logger.setLevel(self.level)
        self._logger.addHandler(handler)

    def log(self, level, message, *args, **kwargs):
        self._logger.log(level, message, *args, **kwargs)



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.logging
    :platform: Unix
    :synopsis: Burp-UI logging module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import logging


def convert_level(verbose):
    # This is already a valid level
    if logging.getLevelName(verbose) != "Level %s" % verbose and (
        not isinstance(verbose, int) or verbose > 0
    ):
        return verbose

    # The debug argument used to be a boolean so we keep supporting this format
    if isinstance(verbose, bool):
        if verbose:
            verbose = logging.DEBUG
        else:
            verbose = logging.CRITICAL
    else:
        levels = [
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ]
        if verbose >= len(levels):
            verbose = len(levels) - 1
        if not verbose:
            verbose = 0
        verbose = levels[verbose]

    return verbose


class Logger(logging.Logger):
    """
    Helper class to share our logging object across the application
    """

    app = None
    _handler = None

    def __init__(self, app=None, name=None, level=logging.NOTSET):
        """
        :param app: Application context
        :type app: flask.Flask

        :param name: Logger name
        :type name: str

        :param level: Default logging level
        :type level: int
        """
        if app and not name:  # pragma: nocover
            name = app.name
        elif not name:
            name = "burp-ui"
        logging.Logger.__init__(self, name, level)
        if app:  # pragma: nocover
            self.init_app(app)

    def init_app(self, app):
        """
        :param app: Application context
        :type app: flask.Flask
        """
        self.app = app
        config = {
            "level": app.config.get("LOG_LEVEL"),
            "logfile": app.config.get("LOG_FILE"),
        }
        self.init_logger(config)

    def init_logger(self, config):
        """
        :param config: Logger configuration
        :type config: dict
        """
        level = config.get("level", None)
        level = self.level if level is None else level
        level = convert_level(level)
        logfile = config.get("logfile")
        if self._handler is not None:
            self.removeHandler(self._handler)
        if logfile:
            handler = logging.FileHandler(logfile)
        else:
            handler = logging.StreamHandler()

        if level > logging.DEBUG:
            LOG_FORMAT = (
                "[%(asctime)s] %(levelname)s in " "%(module)s.%(funcName)s: %(message)s"
            )
        else:
            LOG_FORMAT = (
                "-" * 27
                + "[%(asctime)s]"
                + "-" * 28
                + "\n"
                + "%(levelname)s in %(module)s.%(funcName)s "
                + "[%(pathname)s:%(lineno)d]:\n"
                + "%(message)s\n"
                + "-" * 80
            )

        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))

        self.setLevel(level)
        self.addHandler(handler)
        self._handler = handler


logger = Logger()



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.doc
    :platform: Unix
    :synopsis: Burp-UI parser documentation for Burp1.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .interface import BUIparser


def __(string):
    """dummy function to fake the translation"""
    return string


class Doc(BUIparser):
    """:class:`burpui.misc.parser.doc.Doc` provides a consistent interface
    to parse burp configuration files.

    It implements :class:`burpui.misc.parser.interface.BUIparser`.
    """

    defaults = {
        "address": "",  # IP
        "atime": False,  # bool
        "autoupgrade_dir": "",  # dir
        "ca_burp_ca": "",  # file
        "ca_conf": "",  # file
        "ca_name": "",
        "ca_server_name": "",
        "client_can_delete": True,  # bool
        "client_can_diff": True,  # bool
        "client_can_force_backup": True,  # bool
        "client_can_list": True,  # bool
        "client_can_restore": True,  # bool
        "client_can_verify": True,  # bool
        "client_lockdir": "",
        "clientconfdir": "",  # dir
        "compression": "gzip9",
        "cross_all_filesystems": False,  # bool
        "cross_filesystem": "",
        "daemon": True,  # bool
        "dedup_group": "",
        "directory_tree": True,  # bool
        "directory": "",
        "exclude_comp": "",  # multi
        "exclude_ext": "",  # multi
        "exclude_fs": "",  # multi
        "exclude_regex": "",  # multi
        "exclude": "",  # multi
        "fork": True,  # bool
        "group": "",
        "hard_quota": "",
        "hardlinked_archive": False,  # bool
        "include_ext": "",  # multi
        "include_glob": "",  # multi
        "include_regex": "",  # multi
        "include": "",  # multi
        "keep": [7, 6, 4],  # multi  # int
        "librsync": True,  # bool
        "max_children": 5,  # int
        "max_file_size": "",
        "max_hardlinks": 10000,  # int
        "max_status_children": 5,  # int
        "max_storage_subdirs": 30000,  # int
        "min_file_size": "",
        "mode": "",
        "monitor_browse_cache": False,  # bool
        "network_timeout": 7200,  # int
        "nobackup": "",
        "notify_failure_arg": "",  # multi
        "notify_failure_script": "",  # file
        "notify_success_arg": "",  # multi
        "notify_success_changes_only": False,  # bool
        "notify_success_script": "",  # file
        "notify_success_warnings_only": False,  # bool
        "password": "",
        "password_check": True,  # bool
        "path_length_warn": True,  # bool
        "pidfile": "",
        "port": 4971,  # int
        "protocol": False,  # int
        "ratelimit": False,  # int
        "read_all_blockdevs": False,  # bool
        "read_all_fifos": False,  # bool
        "read_blockdev": "",
        "read_fifo": "",
        "restore_client": "",  # multi
        "scan_problem_raises_error": False,  # bool
        "server_script_arg": "",  # multi
        "server_script_notify": False,  # bool
        "server_script_post_arg": "",  # multi
        "server_script_post_notify": False,  # bool
        "server_script_post_run_on_fail": False,  # bool
        "server_script_post": "",  # file
        "server_script_pre_arg": "",  # multi
        "server_script_pre_notify": False,  # bool
        "server_script_pre": "",  # file
        "server_script": "",  # file
        "soft_quota": "",
        "split_vss": False,  # bool
        "ssl_cert_ca": "",  # file
        "ssl_cert": "",  # file
        "ssl_ciphers": "",
        "ssl_compression": "zlib5",
        "ssl_dhfile": "",  # file
        "ssl_key_password": "",
        "ssl_key": "",  # file
        "ssl_peer_cn": "",
        "status_address": "",  # 127.0.0.1 / ::1
        "status_port": 4972,  # int
        "stdout": True,  # bool
        "strip_vss": False,  # bool
        "syslog": False,  # bool
        "timer_arg": "",  # multi
        "timer_script": "",  # file
        "timestamp_format": "",
        "umask": "0022",  # mode
        "user": "",
        "version_warn": True,  # bool
        "vss_drives": "",
        "working_dir_recovery_method": "",
        "server_can_restore": False,
    }
    placeholders = {
        ".": __("path or glob"),
        "address": __("address"),
        "atime": "0|1",
        "autoupgrade_dir": __("path"),
        "ca_burp_ca": __("path"),
        "ca_conf": __("path"),
        "ca_name": __("name"),
        "ca_server_name": __("name"),
        "client_can_delete": "0|1",
        "client_can_force_backup": "0|1",
        "client_can_list": "0|1",
        "client_can_restore": "0|1",
        "client_can_verify": "0|1",
        "client_lockdir": __("path"),
        "clientconfdir": __("path"),
        "compression": "gzip[0-9]",
        "cross_all_filesystems": "0|1",
        "cross_filesystem": __("path"),
        "daemon": "0|1",
        "dedup_group": __("string"),
        "directory_tree": "0|1",
        "directory": __("path"),
        "exclude_comp": __("extension"),
        "exclude_ext": __("extension"),
        "exclude_fs": __("fstype"),
        "exclude_regex": __("regular expression"),
        "exclude": __("path"),
        "fork": "0|1",
        "group": __("groupname"),
        "hard_quota": "b/Kb/Mb/Gb",
        "hardlinked_archive": "0|1",
        "include_ext": __("extension"),
        "include_regex": __("regular expression"),
        "include": __("path"),
        "include_glob": __("glob"),
        "keep": __("number"),
        "librsync": "0|1",
        "lockfile": __("path"),
        "manual_delete": __("path"),
        "max_children": __("number"),
        "max_file_size": "b/Kb/Mb/Gb",
        "max_hardlinks": __("number"),
        "max_status_children": __("number"),
        "max_storage_subdirs": __("number"),
        "min_file_size": "b/Kb/Mb/Gb",
        "network_timeout": "s",
        "nobackup": __("file name"),
        "notify_failure_arg": __("string"),
        "notify_failure_script": __("path"),
        "notify_success_arg": __("string"),
        "notify_success_changes_only": "0|1",
        "notify_success_script": __("path"),
        "notify_success_warnings_only": "0|1",
        "password": __("password"),
        "password_check": "0|1",
        "pidfile": __("path"),
        "port": __("port number"),
        "ratelimit": "Mb/s",
        "read_all_blockdevs": "0|1",
        "read_all_fifos": "0|1",
        "read_blockdev": __("path"),
        "read_fifo": __("path"),
        "restore_client": __("client"),
        "resume_partial": "0|1",
        "scan_problem_raises_error": "0|1",
        "server_script_arg": __("path"),
        "server_script_notify": "0|1",
        "server_script_post_arg": __("string"),
        "server_script_post_notify": "0|1",
        "server_script_post_run_on_fail": "0|1",
        "server_script_post": __("path"),
        "server_script_pre_arg": __("string"),
        "server_script_pre_notify": "0|1",
        "server_script_pre": __("path"),
        "server_script": __("path"),
        "soft_quota": "b/Kb/Mb/Gb",
        "ssl_cert_ca": __("path"),
        "ssl_cert_password": __("password"),
        "ssl_cert": __("path"),
        "ssl_ciphers": __("cipher list"),
        "ssl_dhfile": __("path"),
        "ssl_key_password": __("password"),
        "ssl_key": __("path"),
        "ssl_peer_cn": __("string"),
        "status_port": __("port number"),
        "stdout": "0|1",
        "strip_vss": "0|1",
        "suplit_vss": "0|1",
        "syslog": "0|1",
        "timer_arg": __("string"),
        "timer_script": __("path"),
        "timestamp_format": __("strftime format"),
        "umask": __("umask"),
        "user": __("username"),
        "version_warn": "0|1",
        "vss_drives": __("list of drive letters"),
        "working_dir_recovery_method": "resume|use|delete",
    }
    values = {
        "compression": ["gzip{0}".format(x) for x in range(0, 10)],
        "mode": ["client", "server"],
        "ssl_compression": ["zlib{0}".format(x) for x in range(0, 10)],
        "status_address": ["127.0.0.1", "::1"],  # 127.0.0.1 / ::1
        "working_dir_recovery_method": ["use", "delete", "resume"],
    }
    files = [
        "ca_burp_ca",
        "ca_conf",
        "notify_failure_script",
        "notify_success_script",
        "server_script_post",
        "server_script_pre",
        "server_script",
        "ssl_cert_ca",
        "ssl_cert",
        "ssl_dhfile",
        "ssl_key",
        "timer_script",
    ]
    advanced_type = {
        "keep": "integer",
    }
    pair_srv = []
    pair_associations = {}
    multi_srv = [
        "exclude_comp",
        "exclude_ext",
        "exclude_fs",
        "exclude_regex",
        "exclude",
        "include_ext",
        "include_glob",
        "include_regex",
        "include",
        "keep",
        "notify_failure_arg",
        "notify_success_arg",
        "restore_client",
        "server_script_arg",
        "server_script_post_arg",
        "server_script_pre_arg",
        "timer_arg",
    ]
    boolean_srv = [
        "atime",
        "client_can_delete",
        "client_can_diff",
        "client_can_force_backup",
        "client_can_list",
        "client_can_restore",
        "client_can_verify",
        "cross_all_filesystems",
        "daemon",
        "directory_tree",
        "fork",
        "hardlinked_archive",
        "librsync",
        "monitor_browse_cache",
        "notify_success_changes_only",
        "notify_success_warnings_only",
        "password_check",
        "path_length_warn",
        "read_all_blockdevs",
        "read_all_fifos",
        "scan_problem_raises_error",
        "server_script_notify",
        "server_script_post_notify",
        "server_script_post_run_on_fail",
        "server_script_pre_notify",
        "split_vss",
        "stdout",
        "strip_vss",
        "syslog",
        "version_warn",
        "server_can_restore",
    ]
    integer_srv = [
        "max_children",
        "max_hardlinks",
        "max_status_children",
        "max_storage_subdirs",
        "network_timeout",
        "port",
        "protocol",
        "ratelimit",
        "status_port",
    ]
    string_srv = [
        "address",
        "ca_burp_ca",
        "ca_conf",
        "ca_name",
        "ca_server_name",
        "client_lockdir",
        "compression",
        "dedup_group",
        "directory",
        "group",
        "hard_quota",
        "mode",
        "notify_failure_script",
        "notify_success_script",
        "pidfile",
        "server_script_post",
        "server_script_pre",
        "server_script",
        "soft_quota",
        "ssl_cert_ca",
        "ssl_cert",
        "ssl_ciphers",
        "ssl_compression",
        "ssl_dhfile",
        "ssl_key_password",
        "ssl_key",
        "status_address",
        "timestamp_format",
        "umask",
        "user",
        "working_dir_recovery_method",
        "min_file_size",
        "max_file_size",
        "cross_filesystem",
        "nobackup",
        "read_fifo",
        "read_blockdev",
        "vss_drives",
    ]
    fields_cli = [
        "atime",
        "client_can_delete",
        "client_can_force_backup",
        "client_can_list",
        "client_can_restore",
        "client_can_verify",
        "compression",
        "cross_all_filesystems",
        "cross_filesystem",
        "dedup_group",
        "directory_tree",
        "directory",
        "exclude_comp",
        "exclude_ext",
        "exclude_fs",
        "exclude_regex",
        "exclude",
        "hard_quota",
        "include_ext",
        "include_regex",
        "include_glob",
        "include",
        "keep",
        "librsync",
        "max_file_size",
        "min_file_size",
        "nobackup",
        "notify_failure_arg",
        "notify_failure_script" "notify_success_arg",
        "notify_success_script",
        "notify_success_warnings_only",
        "password_check",
        "password",
        "path_length_warn",
        "protocol",
        "read_all_blockdevs",
        "read_all_fifos",
        "read_blockdev",
        "read_fifo",
        "restore_client",
        "scan_problem_raises_error",
        "server_script_arg",
        "server_script_notify",
        "server_script_post_arg",
        "server_script_post_notify",
        "server_script_post_run_on_fail",
        "server_script_post",
        "server_script_pre_arg",
        "server_script_pre_notify",
        "server_script_pre",
        "server_script",
        "soft_quota",
        "split_vss",
        "ssl_peer_cn",
        "strip_vss",
        "syslog",
        "timer_arg",
        "timer_script",
        "timestamp_format",
        "version_warn",
        "vss_drives",
        "working_dir_recovery_method",
        "server_can_restore",
    ]
    string_cli = list(set(string_srv) & set(fields_cli))
    string_cli += ["ssl_peer_cn", "password"]
    boolean_cli = list(set(boolean_srv) & set(fields_cli))
    integer_cli = list(set(integer_srv) & set(fields_cli))
    multi_cli = list(set(multi_srv) & set(fields_cli))
    doc = {
        ".": __(
            "Read additional configuration files. On Windows, the glob"
            " is unimplemented - you will need to specify an actual"
            " file."
        ),
        "address": __(
            "Defines the main TCP address that the server listens"
            " on. The default is either '::' or '0.0.0.0',"
            " dependent upon compile time options."
        ),
        "atime": __(
            "This allows you to control whether the client uses"
            " O_NOATIME when opening files and directories. The"
            " default is 0, which enables O_NOATIME. This means that"
            " the client can read files and directories without"
            " updating the access times. However, this is only"
            " possible if you are running as root, or are the owner"
            " of the file or directory. If this is not the case"
            " (perhaps you only have group or world access to the"
            " files), you will get errors until you set atime=1."
            " With atime=1, the access times will be updated on the"
            " files and directories that get backed up."
        ),
        "autoupgrade_dir": __(
            "Path to autoupgrade directory from which"
            " upgrades are downloaded. The option can be"
            " left unset in order not to autoupgrade"
            " clients. Please see docs/autoupgrade.txt in"
            " the source package for more help with this"
            " option."
        ),
        "ca_burp_ca": __(
            "Path to the burp_ca script when using the ca_conf" " option."
        ),
        "ca_conf": __(
            "Path to certificate authority configuration file. The"
            " CA configuration file will usually be"
            " /etc/burp/CA.cnf. The CA directory indicated by"
            " CA.cnf will usually be /etc/burp/CA. If ca_conf is"
            " set and the CA directory does not exist, the server"
            " will create, populate it, and the paths indicated by"
            " ssl_cert_ca, ssl_cert, ssl_key and ssl_dhfile will be"
            " overwritten. For more detailed information on this"
            " and the other ca_* options, please see"
            " docs/burp_ca.txt."
        ),
        "ca_name": __(
            "Name of the CA that the server will generate when"
            " using the ca_conf option."
        ),
        "ca_server_name": __(
            "The name that the server will put into its own"
            " SSL certficates when using the ca_conf"
            " option."
        ),
        "client_can_delete": __(
            "Turn this off to prevent clients from"
            " deleting backups with the '-a D' option."
            " The default is that clients can delete"
            " backups. Restore clients can override this"
            " setting."
        ),
        "client_can_force_backup": __(
            "Turn this off to prevent clients from"
            " forcing backups with the '-a b'"
            " option. Timed backups will still"
            " work. The default is that clients can"
            " force backups."
        ),
        "client_can_list": __(
            "Turn this off to prevent clients from listing"
            " backups with the '-a l' option. The default"
            " is that clients can list backups. Restore"
            " clients can override this setting."
        ),
        "client_can_restore": __(
            "Turn this off to prevent clients from"
            " initiating restores with the '-a r'"
            " option. The default is that clients can"
            " initiate restores. Restore clients can"
            " override this setting."
        ),
        "client_can_verify": __(
            "Turn this off to prevent clients from"
            " initiating a verify job with the '-a v'"
            " option. The default is that clients can"
            " initiate a verify job. Restore clients can"
            " override this setting."
        ),
        "client_lockdir": __(
            "Path to the directory in which to keep"
            " per-client lock files. By default, this is set"
            " to the path given by the 'directory' option."
        ),
        "clientconfdir": __(
            "Path to the directory that contains client" " configuration files."
        ),
        "compression": __(
            "Choose the level of gzip compression for files"
            " stored in backups. Setting 0 or gzip0 turns"
            " compression off. The default is gzip9. This"
            " option can be overridden by the client"
            " configuration files in clientconfdir on the"
            " server."
        ),
        "cross_all_filesystems": __(
            "Allow backups to cross all filesystem" " mountpoints."
        ),
        "cross_filesystem": __(
            "Allow backups to cross a particular" " filesystem mountpoint."
        ),
        "daemon": __("Whether to daemonise. The default is 1."),
        "dedup_group": __(
            "Enables you to group clients together for file"
            " deduplication purposes. For example, you might"
            " want to set 'dedup_group=xp' for each Windows XP"
            " client, and then run the bedup program on a cron"
            " job every other day with the option '-g xp'."
        ),
        "directory_tree": __(
            "When turned on (which is the default) and the"
            " client is on version 1.3.6 or greater, the"
            " structure of the storage directory will mimic"
            " that of the original filesystem on the"
            " client."
        ),
        "directory": __("Path to the directory in which to store backups."),
        "exclude_comp": __(
            "Extensions to exclude from compression. Case"
            " insensitive. You can have multiple exclude"
            " compression lines. For example, set 'gz' to"
            " exclude gzipped files from compression."
        ),
        "exclude_ext": __(
            "Extensions to exclude from the backup. Case"
            " insensitive. You can have multiple exclude"
            " extension lines. For example, set 'vdi' to"
            " exclude VirtualBox disk images."
        ),
        "exclude_fs": __(
            "File systems to exclude from the backup. Case"
            " insensitive. You can have multiple exclude file"
            " system lines. For example, set 'tmpfs' to exclude"
            " tmpfs. Burp has an internal mapping of file system"
            " names to file system IDs. If you know the file"
            " system ID, you can use that instead. For example,"
            " 'exclude_fs = 0x01021994' will also"
            " exclude tmpfs."
        ),
        "exclude_regex": __("Exclude paths that match the regular" " expression."),
        "exclude": __(
            "Path to exclude from the backup. You can have"
            " multiple exclude lines. Use forward slashes '/', not"
            " backslashes '\\' as path delimiters."
        ),
        "fork": __("Whether to fork children. The default is 1."),
        "group": __(
            "Run as a particular group. This can be overridden by"
            " the client configuration files in clientconfdir on the"
            " server."
        ),
        "hard_quota": __(
            "Do not back up the client if the estimated size of"
            " all files is greater than the specified size."
            " Example: 'hard_quota = 100Gb'. Set to 0 (the"
            " default) to have no limit."
        ),
        "hardlinked_archive": __(
            "On the server, defines whether to keep"
            " hardlinked files in the backups, or"
            " whether to generate reverse deltas and"
            " delete the original files. Can be set to"
            " either 0 (off) or 1 (on). Disadvantage:"
            " More disk space will be used Advantage:"
            " Restores will be faster, and since no"
            " reverse deltas need to be generated, the"
            " time and effort the server needs at the"
            " end of a backup is reduced."
        ),
        "include_ext": __(
            "Extensions to include in the backup. Case"
            " insensitive. Nothing else will be included in the"
            " backup. You can have multiple include extension"
            " lines. For example, set 'txt' to include files"
            " that end in '.txt'. You need to specify an"
            " 'include' line so that burp knows where to start"
            " looking."
        ),
        "include_regex": __("Not implemented."),
        "include": __(
            "Path to include in the backup. You can have multiple"
            " include lines. Use forward slashes '/', not"
            " backslashes '\\' as path delimiters."
        ),
        "include_glob": __(
            "Include paths that match the glob expression."
            "For example, '/home/*/Documents' will include"
            " '/home/user1/Documents' and"
            " '/home/user2/Documents' if directories 'user1'"
            " and 'user2' exist in '/home'. The Windows"
            " implementation currently limit the expression to"
            " contain only one '*'."
        ),
        "keep": __(
            "Number of backups to keep. This can be overridden by the"
            " client configuration files in clientconfdir on the"
            " server. Specify multiple 'keep' entries on separate"
            " lines in order to keep multiple periods of backups. For"
            " example, assuming that you are doing a backup a day,"
            " keep=7 keep=4 keep=6 (on separate lines) will keep 7"
            " daily backups, 4 weekly backups (7x4=28), and 6"
            " multiples of 4 weeks (7x4x6=168) - roughly 6 monthly"
            " backups. Effectively, you will be guaranteed to be able"
            " to restore up to 168 days ago, with the number of"
            " available backups exponentially decreasing as you go"
            " back in time to that point. In this example, every 7th"
            " backup will be hardlinked to allow burp to safely delete"
            " intermediate backups when necessary. You can have as"
            " many 'keep' lines as you like, as long as they don't"
            " exceed 52560000 when multiplied together. That is, a"
            " backup every minute for 100 years."
        ),
        "librsync": __(
            "When set to 0, delta differencing will not take"
            " place. That is, when a file changes, the server will"
            " request the whole new file. The default is 1. This"
            " option can be overridden by the client configuration"
            " files in clientconfdir on the server."
        ),
        "lockfile": __(
            "Path to the lockfile that ensures that two server"
            " processes cannot run simultaneously."
        ),
        "manual_delete": __(
            "If a path is given, the server will move"
            " directories to be deleted into the directory"
            " specified by the path, but will not actually"
            " delete them. The path must be on the same file"
            " system as the backup storage. The idea is that"
            " a busy server may be configured to run the"
            " deletions outside of the backup timebands, when"
            " the server is less busy, via a cron job. The"
            " default is unset, which means that the server"
            " will automatically delete the directories at"
            " the end of a backup. This option can be"
            " overridden by the client configuration files in"
            " clientconfdir on the server."
        ),
        "max_children": __(
            "Defines the number of child processes to fork"
            " (the number of clients that can simultaneously"
            " connect. The default is 5."
        ),
        "max_file_size": __(
            "Do not back up files that are greater than the"
            " specified size. Example: 'max_file_size ="
            " 10Mb'. Set to 0 (the default) to have no"
            " limit."
        ),
        "max_hardlinks": __(
            "On the server, the number of times that a single"
            " file can be hardlinked. The bedup program also"
            " obeys this setting. The default is 10000."
        ),
        "max_status_children": __(
            "Defines the number of status child"
            " processes to fork (the number of status"
            " clients that can simultaneously connect."
            " The default is 5."
        ),
        "max_storage_subdirs": __(
            "Defines the number of subdirectories in"
            " the data storage areas. The maximum number"
            " of subdirectories that ext3 allows is"
            " 32000. If you do not set this option, it"
            " defaults to 30000."
        ),
        "min_file_size": __(
            "Do not back up files that are less than the"
            " specified size. Example: 'min_file_size ="
            " 10Mb'. Set to 0 (the default) to have no"
            " limit."
        ),
        "mode": __("Required to run in server mode."),
        "monitor_browse_cache": __(
            "Whether or not the server should cache"
            " the directory tree when a monitor client"
            " is browsing. <br/>Advantage: browsing is"
            " faster. </br>Disadvantage: more memory is"
            " used."
        ),
        "network_timeout": __(
            "Set the network timeout in seconds. If no"
            " data is sent or received over a period of"
            " this length, burp will give up. The default"
            " is 7200 seconds (2 hours)."
        ),
        "nobackup": __(
            "If this file system entry exists, the directory"
            " containing it will not be backed up."
        ),
        "notify_failure_arg": __(
            "The same as notify_success_arg, but for" " backups that failed."
        ),
        "notify_failure_script": __(
            "The same as notify_success_script, but" " for backups that failed."
        ),
        "notify_success_arg": __(
            "A user-definable argument to the notify"
            " success script. You can have many of"
            " these. The notify_success_arg options can"
            " be overridden by the client configuration"
            " files in clientconfdir on the server."
        ),
        "notify_success_script": __(
            "Path to the script to run when a backup"
            " succeeds. User arguments are appended"
            " after the first five reserved"
            " arguments. An example notify script is"
            " provided. The notify_success_script"
            " option can be overridden by the client"
            " configuration files in clientconfdir on"
            " the server."
        ),
        "notify_success_warnings_only": __(
            "Set to 1 to send success"
            " notifications when there were"
            " warnings. If this and"
            " notify_success_changes_only are"
            " not turned on, success"
            " notifications are always sent."
        ),
        "password": __("Defines the password to send to the server."),
        "password_check": __(
            "Allows you to turn client password checking on"
            " or off. The default is on. SSL certificates"
            " will still be checked if you turn passwords"
            " off. This option can be overridden by the"
            " client configuration files in clientconfdir on"
            " the server."
        ),
        "path_length_warn": __(
            "When this is on, which is the default, a"
            " warning will be issued when the client sends"
            " a path that is too long to replicate in the"
            " storage area tree structure. The file will"
            " still be saved in a numbered file outside of"
            " the tree structure, regardless of the"
            " setting of this option. This option can be"
            " overridden by the client configuration files"
            " in clientconfdir on the server."
        ),
        "pidfile": __("Synonym for lockfile."),
        "port": __("Defines the main TCP port that the server listens on."),
        "protocol": __(
            "Choose which style of backups and restores to use. 0"
            " (the default) automatically decides based on the"
            " server version and which protocol is set on the"
            " server side. 1 forces protocol1 style (file level"
            " granularity with a pseudo mirrored storage on the"
            " server and optional rsync). 2 forces protocol2 style"
            " (inline deduplication with variable length blocks)."
            " If you choose a forced setting, it will be an error"
            " if the server also chooses a forced setting."
        ),
        "ratelimit": __(
            "Set the network send rate limit, in Mb/s. If this"
            " option is not given, burp will send data as fast as"
            " it can."
        ),
        "read_all_blockdevs": __(
            "Open all block devices for reading and"
            " back up the contents as if they were"
            " regular files."
        ),
        "read_all_fifos": __(
            "Open all fifos for reading and back up the"
            " contents as if they were regular files."
        ),
        "read_blockdev": __(
            "Do not back up the given block device itself,"
            " but open it for reading and back up the"
            " contents as if it were a regular file."
        ),
        "read_fifo": __(
            "Do not back up the given fifo itself, but open it"
            " for reading and back up the contents as if it were"
            " a regular file."
        ),
        "restore_client": __(
            "A client that is permitted to list, verify,"
            " restore and delete files belonging to any"
            " other client. You may specify multiple"
            " restore_clients. If this is too permissive,"
            " you may set a restore_client for individual"
            " original clients in the individual"
            " clientconfdir files. Note that restoring a"
            " backup from a Windows computer onto a Linux"
            " computer will currently leave the VSS headers"
            " in place at the beginning of each file. This"
            " will be addressed in a future version of"
            " burp."
        ),
        "resume_partial": __(
            "Turn this on to enable 'resume partial' code."
            " Requires 'working_dir_recovery_method=resume'."
            " When resuming an interrupted transfer of a"
            " single file, it attempts to use previously"
            " transferred blocks of that file in order to be"
            " more efficient. However, situations have been"
            " reported where the file on the server side"
            " just gets bigger forever, so this feature now"
            " defaults to being turned off."
        ),
        "scan_problem_raises_error": __(
            "When enabled, this causes problems"
            " in the phase1 scan (such as an"
            " 'include' being missing) to be"
            " treated as fatal errors. The"
            " default is off."
        ),
        "server_script_arg": __(
            "Goes with server_script and overrides"
            " server_script_pre_arg and"
            " server_script_post_arg."
        ),
        "server_script_notify": __(
            "Turn on to send a notification emails"
            " when the server pre and post scripts"
            " return non-zero. The output of the"
            " script will be included it the email."
            " The default is off. Requires the"
            " notify_failure options to be set."
        ),
        "server_script_post_arg": __(
            "A user-definable argument to the"
            " server post script. You can have many"
            " of these."
        ),
        "server_script_post_notify": __(
            "Turn on to send a notification"
            " email when the server post script"
            " returns non-zero. The output of the"
            " script will be included in the"
            " email. The default is off. Requires"
            " the notify_failure options to be"
            " set."
        ),
        "server_script_post_run_on_fail": __(
            "If this is set to 1,"
            " server_script_post will always"
            " be run. The default is 0,"
            " which means that if the task"
            " asked for by the client fails,"
            " server_script_post will not be"
            " run."
        ),
        "server_script_post": __(
            "Path to a script to run on the server"
            " before the client disconnects. The"
            " arguments to it are 'post', '(client"
            " command)', '(client name), '(0 or 1 for"
            " success or failure)', '(timer script exit"
            " code)', and"
            " then arguments defined by"
            " server_script_post_arg. This command and"
            " related options can be overriddden by the"
            " client configuration files in"
            " clientconfdir on the server."
        ),
        "server_script_pre_arg": __(
            "A user-definable argument to the server"
            " pre script. You can have many of"
            " these."
        ),
        "server_script_pre_notify": __(
            "Turn on to send a notification email"
            " when the server pre script returns"
            " non-zero. The output of the script"
            " will be included in the email. The"
            " default is off. Most people will not"
            " want this turned on because clients"
            " usually contact the server at 20"
            " minute intervals and this could"
            " cause a lot of emails to be"
            " generated. Requires the"
            " notify_failure options to be set."
        ),
        "server_script_pre": __(
            "Path to a script to run on the server after"
            " each successfully authenticated connection"
            " but before any work is carried out. The"
            " arguments to it are 'pre', '(client"
            " command)', '(client name)', '(0 or 1 for"
            " success or failure)', '(timer script exit"
            " code)', and"
            " then arguments defined by"
            " server_script_pre_arg. If the script"
            " returns non-zero, the task asked for by the"
            " client will not be run. This command and"
            " related options can be overriddden by the"
            " client configuration files in clientconfdir"
            " on the server."
        ),
        "server_script": __(
            "You can use this to save space in your config"
            " file when you want to run the same server"
            " script twice. It overrides server_script_pre"
            " and server_script_post. This command and"
            " related options can be overriddden by the"
            " client configuration files in clientconfdir on"
            " the server."
        ),
        "soft_quota": __(
            "A warning will be issued when the estimated size"
            " of all files is greater than the specified size"
            " and smaller than hard_quota. Example: 'soft_quota"
            " = 95Gb'. Set to 0 (the default) to have no"
            " warning."
        ),
        "split_vss": __(
            "When backing up Windows computers with burp"
            " protocol 1, this option allows you to save the VSS"
            " header data separate from the file data. The"
            " default is off, which means that the VSS header"
            " data is saved prepended to the file data."
        ),
        "ssl_cert_ca": __(
            "The path to the SSL CA certificate. This file"
            " will probably be the same on both the server and"
            " the client. The file should contain just the"
            " certificate in PEM format. For more information"
            " on this, and the other ssl_* options, please see"
            " <a href='http://burp.grke.org/docs/burp_ca.html'>"
            " docs/burp_ca.txt</a>."
        ),
        "ssl_cert_password": __("Synonym for ssl_key_password."),
        "ssl_cert": __(
            "The path to the server SSL certificate. It works for"
            " me when the file contains the concatenation of the"
            " certificate and private key in PEM format."
        ),
        "ssl_ciphers": __("Allowed SSL ciphers. See openssl ciphers for" " details."),
        "ssl_compression": __(
            "Choose the level of zlib compression over"
            " SSL. Setting 0 or zlib0 turnsSSL compression"
            " off. Setting non-zero gives zlib5 compression"
            " (it is not currently possible for openssl to"
            " set any other level). The default is 5."
            " 'gzip' is a synonym of 'zlib'.is a synonym of"
            " 'zlib'."
        ),
        "ssl_dhfile": __(
            "Path to Diffie-Hellman parameter file. To generate"
            " one with openssl, use a command like this: openssl"
            " dhparam -out dhfile.pem -5 1024"
        ),
        "ssl_key_password": __("The SSL key password."),
        "ssl_key": __("The path to the server SSL private key in PEM" " format."),
        "ssl_peer_cn": __(
            "Must match the common name in the SSL certificate"
            " that the server giveswhen it connects. If"
            " ssl_peer_cn is not set, the server name will be"
            " used instead."
        ),
        "status_address": __(
            "Defines the main TCP address that the server"
            " listens on for status requests. The default is"
            " either '::1' or '127.0.0.1', dependent upon"
            " compile time options."
        ),
        "status_port": __(
            "Defines the TCP port that the server listens on" " for status requests."
        ),
        "stdout": __("Log to stdout. Defaults to on."),
        "strip_vss": __(
            "When backing up Windows computers with burp"
            " protocol 1, this option allows you to prevent the"
            " VSS header data being backed up. The default is"
            " off. To restore a backup that has no VSS"
            " information on Windows, you need to give the client"
            " the '-x' command line option."
        ),
        "syslog": __("Log to syslog. Defaults to off."),
        "timer_arg": __(
            "A user-definable argument to the timer script."
            "You can have many of these. The timer_arg options"
            " can be overridden by the client configuration files"
            " in clientconfdir on the server."
        ),
        "timer_script": __(
            "Path to the script to run when a client connects"
            " with the timed backup option. If the script"
            " exits with code 0, a backup will run. The first"
            " two arguments are the client name and the path"
            " to the 'current' storage directory. The next"
            " three arguments are reserved, and user arguments"
            " are appended after that. An example timer script"
            " is provided. The timer_script option can be"
            " overridden by the client configuration files in"
            " clientconfdir on the server."
        ),
        "timestamp_format": __(
            "This allows you to tweak the format of the"
            " timestamps of individual backups. See 'man"
            " strftime' to see available substitutions."
            " If this option is unset, burp uses"
            ' "%Y-%m-%d %H:%M:%S".'
        ),
        "umask": __("Set the file creation umask. Default is 0022."),
        "user": __(
            "Run as a particular user. This can be overridden by the"
            " client configuration files in clientconfdir on the"
            " server."
        ),
        "version_warn": __(
            "When this is on, which is the default, a warning"
            " will be issued when the client version does not"
            " match the server version. This option can be"
            " overridden by the client configuration files in"
            " clientconfdir on the server."
        ),
        "vss_drives": __(
            "When backing up Windows computers, this option"
            " allows you to specify which drives have VSS"
            " snapshots taken of them. If you omit this option,"
            " burp will automatically decide based on the"
            " 'include' options. If you want no drives to have"
            " snapshots taken of them, you can specify '0'."
        ),
        "working_dir_recovery_method": __(
            "This option tells the server what"
            " to do when it finds the working"
            " directory of an interrupted"
            " backup (perhaps somebody pulled"
            " the plug on the server, or"
            " something). This can be"
            " overridden by the client"
            " configurations files in"
            " clientconfdir on the server."
            " Options are... <ul><li>delete:"
            " Just delete the old working"
            " directory.</li><li>use: Convert"
            " the working directory into a"
            " complete backup.</li><li>resume:"
            " Simply continue the previous"
            " backup from the point at which it"
            " left off, at file granularity."
            " NOTE: If the client has changed"
            " its include/exclude configuration"
            " since the backup was interrupted,"
            " the recovery method will"
            " automatically switch to 'use'."
            " </li></ul>"
        ),
    }

    @property
    def all(self):
        return sorted(
            list(
                set(self.boolean_srv)
                | set(self.files)
                | set(self.integer_srv)
                | set(self.multi_srv)
                | set(self.string_srv)
                | set(self.boolean_cli)
                | set(self.integer_cli)
                | set(self.multi_cli)
                | set(self.string_cli)
            )
        )



# -*- coding: utf8 -*-
"""
.. module:: burpui.utils
    :platform: Unix
    :synopsis: Burp-UI utils module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import math
import string
import sys
import datetime
import zipfile
import tarfile

from uuid import UUID

NOTIF_OK = 0
NOTIF_WARN = 1
NOTIF_ERROR = 2
NOTIF_INFO = 3


class human_readable(int):
    """define a human_readable class to allow custom formatting
    format specifiers supported :
        em : formats the size as bits in IEC format i.e. 1024 bits (128 bytes) = 1Kib
        eM : formats the size as Bytes in IEC format i.e. 1024 bytes = 1KiB
        sm : formats the size as bits in SI format i.e. 1000 bits = 1kb
        sM : formats the size as bytes in SI format i.e. 1000 bytes = 1KB
        cm : format the size as bit in the common format i.e. 1024 bits (128 bytes) = 1Kb
        cM : format the size as bytes in the common format i.e. 1024 bytes = 1KB

    code from: http://code.activestate.com/recipes/578323-human-readable-filememory-sizes-v2/
    """

    def __format__(self, fmt):  # pragma: no cover
        # is it an empty format or not a special format for the size class
        if fmt == "" or fmt[-2:].lower() not in ["em", "sm", "cm"]:
            if fmt[-1].lower() in ["b", "c", "d", "o", "x", "n", "e", "f", "g", "%"]:
                # Numeric format.
                return int(self).__format__(fmt)
            else:
                return str(self).__format__(fmt)

        if sys.version_info >= (3, 0):
            chars = string.ascii_lowercase
        else:
            chars = string.lowercase
        # work out the scale, suffix and base
        factor, suffix = (8, "b") if fmt[-1] in chars else (1, "B")
        base = 1024 if fmt[-2] in ["e", "c"] else 1000

        # Add the i for the IEC format
        suffix = "i" + suffix if fmt[-2] == "e" else suffix

        mult = ["", "K", "M", "G", "T", "P"]

        val = float(self) * factor
        i = 0 if val < 1 else int(math.log(val, base)) + 1
        v = val / math.pow(base, i)
        v, i = (v, i) if v > 0.5 else (v * base, i - 1)

        # Identify if there is a width and extract it
        width = "" if fmt.find(".") == -1 else fmt[: fmt.index(".")]
        precis = fmt[:-2] if width == "" else fmt[fmt.index(".") : -2]

        # do the precision bit first, so width/alignment works with the suffix
        if float(self) == 0:
            return "{0:{1}f}".format(v, precis)
        t = ("{0:{1}f} " + mult[i] + suffix).format(v, precis)

        return "{0:{1}}".format(t, width) if width != "" else t


class BUIcompress:
    """Provides a context to generate any kind of archive supported by
    burp-ui
    """

    def __init__(self, name, archive, zip64=False):  # pragma: no cover
        self.name = name
        self.archive = archive
        self.zip64 = zip64

    def __enter__(self):
        self.arch = None
        if self.archive == "zip":
            self.arch = zipfile.ZipFile(
                self.name,
                mode="w",
                compression=zipfile.ZIP_DEFLATED,
                allowZip64=self.zip64,
            )
        elif self.archive == "tar.gz":
            self.arch = tarfile.open(self.name, "w:gz")
        elif self.archive == "tar.bz2":
            self.arch = tarfile.open(self.name, "w:bz2")
        return self

    def __exit__(self, type, value, traceback):
        self.arch.close()

    def append(self, path, arcname):
        if self.archive == "zip":
            if os.path.islink(path):
                # This is a symlink, we virtually create one in memory
                # because zipfile does not seem to support them natively
                vfile = zipfile.ZipInfo()
                vfile.filename = arcname  # That's the name of the actual file
                vfile.external_attr |= 0o120000 << int(16)  # symlink file type
                vfile.compress_type = zipfile.ZIP_STORED
                # os.readlink gives us the target of the symlink
                self.arch.writestr(vfile, os.readlink(path))
            else:
                self.arch.write(path, arcname)
        elif self.archive in ["tar.gz", "tar.bz2"]:
            self.arch.add(path, arcname=arcname, recursive=False)


def is_uuid(string):
    """Tells if a given string is a UUID"""
    try:
        val = UUID(string, version=4)
    except ValueError:
        return False

    return val.hex == string.replace("-", "")


def lookup_file(name=None, guess=True, directory=False, check=True):
    if name and isinstance(name, str):
        if os.path.isfile(name) or name == "/dev/null":
            return name
        elif directory and os.path.isdir(name):
            return name
        elif not guess:
            if check:
                raise IOError("File not found: '{}'".format(name))
            return name
    if name and isinstance(name, str):
        names = [name]
    elif name:
        names = name
    else:
        names = ["burpui.cfg", "burpui.sample.cfg"]
    roots = [
        "",
        "share/burpui",
        "/etc/burp",
        os.path.join(
            sys.prefix,
            "share",
            "burpui",
        ),
        os.path.join(
            sys.prefix,
            "local",
            "share",
            "burpui",
        ),
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "..",
            "..",
            "..",
            "..",
            "share",
            "burpui",
        ),
    ]
    prefixes = ["", "etc"]
    for filename in names:
        for root in roots:
            for prefix in prefixes:
                tmp = os.path.join(root, prefix, filename)
                if directory and os.path.isdir(tmp):
                    return tmp
                elif os.path.isfile(tmp):
                    return tmp


def utc_to_local(timestamp):
    try:
        import arrow
        from tzlocal import get_localzone

        # 1487607525 -> 1487611125
        utc = arrow.get(datetime.datetime.fromtimestamp(timestamp))
        local = utc.to(str(get_localzone()))
        return local.int_timestamp
    except (TypeError, arrow.parser.ParserError, ImportError):
        return timestamp


def __(string):
    """dummy function to fake the translation"""
    return string


class ReverseProxied(object):
    """Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    In nginx:

    ::

        location /myprefix {
            proxy_pass http://192.168.0.1:5001;
            proxy_set_header Host $host;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Scheme $scheme;
            proxy_set_header X-Script-Name /myprefix;
        }


    In Apache:

    ::

        <Location /myprefix>
            ProxyPass http://192.168.0.1:5001
            ProxyPassReverse http://192.168.0.1:5001
            RequestHeader set X-Script-Name /myprefix
        </Location>


    :param wsgi_app: the WSGI application

    Inspired by: http://flask.pocoo.org/snippets/35/
    """

    def __init__(self, wsgi_app, app):
        self.wsgi_app = wsgi_app
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME", self.app.config["BUI_PREFIX"])
        if script_name:
            if script_name.startswith("/"):
                environ["SCRIPT_NAME"] = script_name
                path_info = environ["PATH_INFO"]
                if path_info.startswith(script_name):
                    environ["PATH_INFO"] = path_info[len(script_name) :]
            else:
                self.app.warning("'prefix' must start with a '/'!")

        return self.wsgi_app(environ, start_response)


def make_list(data):
    """Ensure data is a list

    :param data: data to check

    :returns: data as list
    """
    if isinstance(data, list):
        return data
    if data is None:
        return []
    return list(data) if not isinstance(data, str) else [data]



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.auth.interface
    :platform: Unix
    :synopsis: Burp-UI authentication interface.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask_login import UserMixin, AnonymousUserMixin
from abc import ABCMeta, abstractmethod, abstractproperty

from ...tools.logging import logger


class BUIloader:
    logger = logger


class BUIhandler(object, metaclass=ABCMeta):
    """The :class:`burpui.misc.auth.interface.BUIhandler` class maintains a list
    of ``Burp-UI`` users.

    :param app: Instance of the app we are running in
    :type app: :class:`burpui.engines.server.BUIServer`
    """

    priority = 0

    name = None

    add_user = False
    del_user = False
    change_password = False

    preload_users = True

    @abstractmethod
    def user(self, name=None, refresh=False):
        """The :func:`burpui.misc.auth.interface.BUIhandler.user` function
        returns the :class:`flask_login:flask_login.UserMixin` object
        corresponding to the given user name.

        :param name: Name of the user
        :type name: str

        :param refresh: Whether we need to re-create a fresh user or not
        :type refresh: bool

        :returns: :class:`burpui.misc.auth.interface.BUIuser`
        """
        return AnonymousUserMixin()

    def remove(self, name):
        """The :func:`burpui.misc.auth.interface.BUIhandler.remove` function
        allows to remove a user from the cache.

        :param name: Name of the user to remove
        :type name: str
        """
        pass

    @abstractproperty
    @property
    def loader(self):
        return None


class BUIuser(UserMixin, metaclass=ABCMeta):
    """The :class:`burpui.misc.auth.interface.BUIuser` class extends the
    :class:`flask_login:flask_login.UserMixin` class.
    """

    backend = None
    admin = True
    moderator = True

    @abstractmethod
    def login(self, passwd=None):
        """The :func:`burpui.misc.auth.interface.BUIuser.login` function
        checks if the profided username and password match.

        :param passwd: Password
        :type passwd: str

        :returns: True if the name and password match, otherwise False
        """
        return False  # pragma: no cover

    @property
    def is_active(self):
        """
        :returns: True if user is active, otherwise False
        """
        return self.active

    @property
    def is_authenticated(self):
        """
        :returns: True if a user is authenticated, otherwise False
        """
        return self.authenticated

    @property
    def is_admin(self):
        """
        If no ACL engine is loaded, every logged-in user will be granted admin
        rights
        :returns: True if the user is admin, otherwise False
        """
        return self.admin

    @property
    def is_moderator(self):
        """
        If no ACL engine is loaded, every logged-in user will be granted
        moderator rights
        :returns: True if the user is moderator, otherwise False
        """
        return self.moderator

    def __str__(self):
        msg = UserMixin.__str__(self)
        return "{} (id: {}, name: {}, backend: {}, admin: {}, moderator: {}, authenticated: {}, active: {})".format(
            msg,
            self.get_id(),
            self.name,
            self.backend,
            self.is_admin,
            self.is_moderator,
            self.is_authenticated,
            self.is_active,
        )



# -*- coding: utf8 -*-
"""
.. module:: burpui.engines.monitor
    :platform: Unix
    :synopsis: Burp-UI monitor pool module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import ssl
import trio
import json
import math
import struct
import logging
import datetime

from functools import partial
from itertools import count
from async_generator import asynccontextmanager

from ..exceptions import BUIserverException
from ..misc.backend.utils.burp2 import Monitor
from ..config import config
from .._compat import to_bytes, to_unicode
from ..tools.logging import logger
from ..desc import __version__


CONNECTION_COUNTER = count()


BUI_DEFAULTS = {
    "Global": {
        "port": 11111,
        "bind": "::1",
        "ssl": False,
        "sslcert": "",
        "sslkey": "",
        "password": "password123456",
        "pool": 5,
    },
}


class Pool:
    def __init__(self, pool_size):
        self._size = pool_size
        self.send_channel, self.receive_channel = trio.open_memory_channel(pool_size)

    @property
    def size(self):
        return self._size

    @property
    def stats(self):
        return self.send_channel.statistics()

    async def put(self, data):
        await self.send_channel.send(data)

    async def get(self):
        return await self.receive_channel.receive()

    def empty(self):
        stats = self.stats
        if self.size != 0 and stats.current_buffer_used == 0:
            return True
        return False

    def full(self):
        stats = self.stats
        max_buffer = self.size
        if max_buffer > 0 and max_buffer != math.inf:
            return stats.current_buffer_used == max_buffer
        return False

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.send_channel.aclose()
        await self.receive_channel.aclose()


class MonitorPool:
    logger = logger

    # cache status results
    _status_cache = {}
    _last_status_cleanup = datetime.datetime.now()
    _time_to_cache = datetime.timedelta(seconds=5)

    def __init__(self, conf=None, level=0, logfile=None):
        level = level or 0
        self.logger.init_logger(config=dict(level=level, logfile=logfile))
        lvl = logging.getLevelName(self.logger.getEffectiveLevel())
        self.logger.info(f"conf: {conf}")
        self.logger.info(f"level: {lvl}")
        if not conf:
            raise IOError("No configuration file found")

        # Raise exception if errors are encountered during parsing
        self.conf = config
        self.conf.parse(conf, BUI_DEFAULTS)
        self.conf.default_section("Global")
        self.port = self.conf.safe_get("port", "integer")
        self.bind = self.conf.safe_get("bind")
        self.ssl = self.conf.safe_get("ssl", "boolean")
        self.sslcert = self.conf.safe_get("sslcert")
        self.sslkey = self.conf.safe_get("sslkey")
        self.password = self.conf.safe_get("password")
        self.pool_size = self.conf.safe_get("pool", "integer")

        self.burpbin = self.conf.safe_get("burpbin", section="Burp")
        self.bconfcli = self.conf.safe_get("bconfcli", section="Burp")
        self.timeout = self.conf.safe_get(
            "timeout", "integer", section="Burp", defaults=15
        )

        self.conf.setdefault("BUI_MONITOR", True)

        self.pool = Pool(self.pool_size)

    def _ssl_context(self):
        if not self.ssl:
            return None
        ctx = ssl.SSLContext()
        ctx.load_cert_chain(self.sslcert, self.sslkey)
        return ctx

    def _cleanup_cache(self):
        now = datetime.datetime.now()
        if now - self._last_status_cleanup > self._time_to_cache:
            self._status_cache.clear()
            self._last_status_cleanup = now

    async def receive_all(self, stream: trio.abc.Stream, length=1024, bsize=None):
        buf = b""
        bsize = bsize if bsize is not None else 1024
        bsize = min(bsize, length)
        received = 0
        tries = 0
        while received < length:
            newbuf = await stream.receive_some(bsize)
            if not newbuf:
                # 3 successive read failure => raise exception
                if tries > 3:
                    raise IOError("Unable to read full response")
                tries += 1
                await trio.sleep(0.1)
                continue
            # reset counter
            tries = 0
            buf += newbuf
            received += len(newbuf)
        return buf

    @asynccontextmanager
    async def get_mon(self, ident) -> Monitor:
        self.logger.info(f"{ident} - Waiting for a monitor...")
        t1 = trio.current_time()
        mon = await self.pool.get()  # type: Monitor
        t2 = trio.current_time()
        t = t2 - t1
        self.logger.info(f"{ident} - Waited {t:.3f}s")
        yield mon
        self.logger.info(f"{ident} - Releasing monitor")
        await self.pool.put(mon)

    async def handle(self, server_stream: trio.abc.Stream):
        try:
            ident = next(CONNECTION_COUNTER)
            self.logger.info(f"{ident} - handle_request: started")
            t0 = trio.current_time()
            lengthbuf = await server_stream.receive_some(8)
            if not lengthbuf:
                return
            (length,) = struct.unpack("!Q", lengthbuf)
            data = await self.receive_all(server_stream, length)
            self.logger.info(f"{ident} - recv: {data!r}")
            txt = to_unicode(data)
            if txt == "RE":
                return
            req = json.loads(txt)
            if req["password"] != self.password:
                self.logger.warning(f"{ident} -----> Wrong Password <-----")
                await server_stream.send_all(b"KO")
                return
            try:
                func = req.get("func")
                if func == "monitor_version":
                    response = __version__
                elif func in [
                    "client_version",
                    "server_version",
                    "batch_list_supported",
                ]:
                    async with self.get_mon(ident) as mon:
                        response = getattr(mon, func, "")
                        if func in ["batch_list_supported"]:
                            response = json.dumps(response)
                elif func == "statistics":
                    tmp = []
                    res = {
                        "alive": False,
                        "server_version": "unknown",
                        "client_version": "unknown",
                    }
                    while not res["alive"] and len(tmp) < self.pool.size:
                        mon = await self.pool.get()
                        tmp.append(mon)
                        if mon.alive:
                            res = {
                                "alive": True,
                                "server_version": getattr(mon, "server_version", ""),
                                "client_version": getattr(mon, "client_version", ""),
                            }
                            break
                        await trio.sleep(0.5)
                    for mon in tmp:
                        await self.pool.put(mon)
                    response = json.dumps(res)
                else:
                    query = req["query"]
                    cache = req.get("cache", True)

                    self._cleanup_cache()
                    # return cached results
                    if cache and query in self._status_cache:
                        response = self._status_cache[query]
                    else:
                        async with self.get_mon(ident) as mon:
                            wrap = partial(
                                mon.status,
                                query,
                                timeout=self.timeout,
                                cache=False,
                                raw=True,
                            )
                            response = await trio.to_thread.run_sync(wrap)

                        if cache:
                            self._status_cache[query] = response
                self.logger.debug(f"{ident} - Sending: {response}")
                if response:
                    await server_stream.send_all(b"OK")
                else:
                    await server_stream.send_all(b"KO")
            except BUIserverException as exc:
                await server_stream.send_all(b"ER")
                response = str(exc)
                self.logger.error(response, exc_info=exc)
                self.logger.warning(f"Forwarding Exception: {response}")

            if response:
                response = to_bytes(response)
                await server_stream.send_all(struct.pack("!Q", len(response)))
                await server_stream.send_all(response)

            t3 = trio.current_time()
            t = t3 - t0
            self.logger.info(f"{ident} - Completed in {t:.3f}s")
        except Exception as exc:
            self.logger.error(f"Unexpected error: {exc}")
            response = str(exc)
            self.logger.error(response, exc_info=exc)
            try:
                await server_stream.send_all(b"ER")
                self.logger.warning(f"Forwarding Exception: {response}")

                response = to_bytes(response)
                await server_stream.send_all(struct.pack("!Q", len(response)))
                await server_stream.send_all(response)
            except trio.BrokenResourceError:
                # Broken Pipe, we cannot forward the error
                pass

    async def launch_monitor(self, id):
        self.logger.info(f"Starting client n°{id}")
        try:
            if self.pool.full():
                self.logger.warning("pool full!")
                return
            mon = Monitor(self.burpbin, self.bconfcli, timeout=self.timeout, ident=id)
            # warm up monitor
            await trio.to_thread.run_sync(mon.status)
            await self.pool.put(mon)
        except (BUIserverException, OSError):
            pass

    async def cleanup_monitor(self):
        while not self.monitor_pool.empty():
            self.logger.info("killing proc")
            mon = await self.monitor_pool.get()  # noqa
            del mon

    async def fill_pool(self):
        self.logger.info("Starting clients...")
        async with trio.open_nursery() as nursery:
            for i in range(self.pool_size):
                nursery.start_soon(self.launch_monitor, i + 1)

    async def _run(self):
        self.logger.info(f"Ready to serve requests on {self.bind}:{self.port}")
        ctx = self._ssl_context()
        if ctx:
            await trio.serve_ssl_over_tcp(self.handle, self.port, ctx, host=self.bind)
        else:
            await trio.serve_tcp(self.handle, self.port, host=self.bind)

    async def run(self):
        async with self.pool:
            try:
                async with trio.open_nursery() as nursery:
                    # listen to connections as soon as possible
                    nursery.start_soon(self._run)
                    # in parallel we start to populate the pool
                    nursery.start_soon(self.fill_pool)
            except KeyboardInterrupt:
                pass



"""hidden hosts management

Revision ID: 4445080944ee
Revises: 695dcbd29d4f
Create Date: 2018-10-03 11:47:20.028686

"""

# revision identifiers, used by Alembic.
revision = "4445080944ee"
down_revision = "695dcbd29d4f"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "hidden",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user", sa.String(length=256), nullable=False),
        sa.Column("client", sa.String(length=4096), nullable=True),
        sa.Column("server", sa.String(length=4096), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("hidden")
    # ### end Alembic commands ###



# -*- coding: utf8 -*-
"""
.. module:: burpui.server
    :platform: Unix
    :synopsis: Burp-UI server module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re
import sys
import logging  # noqa
import warnings

from ..tools.logging import logger
from ..misc.auth.handler import UserAuthHandler
from ..misc.acl.handler import ACLloader
from ..misc.audit.handler import BUIauditLoader
from ..config import config
from ..plugins import PluginManager

from datetime import timedelta
from flask import Flask


BUI_DEFAULTS = {
    "Global": {
        "port": 0,
        "bind": "",
        "ssl": False,
        "sslcert": "",
        "sslkey": "",
        "backend": "burp2",
        "auth": ["basic"],
        "acl": ["none"],
        "audit": ["none"],
        "prefix": "",
        "plugins": [],
        "demo": False,
        "dsn": "",
        "piwik_url": "",
        "piwik_script": "piwik.php",
        "piwik_id": 0,
    },
    "UI": {
        "refresh": 180,
        "liverefresh": 5,
        "ignore_labels": ["color:.*"],
        "format_labels": [r"s/^os:\s*//"],
        "default_strip": 0,
    },
    "Security": {
        "scookie": True,
        "appsecret": "random",
        "cookietime": 14,
        "sessiontime": 5,
    },
    "Production": {
        "storage": "",
        "session": "",
        "cache": "",
        "redis": "",
        "celery": False,
        "database": "",
        "limiter": False,
        "ratio": "60/minute",
        "num_proxies": 0,
        "proxy_fix_args": "{'x_proto': {num_proxies}, 'x_for': {num_proxies}, 'x_host': {num_proxies}, 'x_prefix': {num_proxies}}",
    },
    "WebSocket": {
        "enabled": True,
        "embedded": False,
        "broker": "redis",
        "url": "",
        "debug": False,
    },
    "Experimental": {
        "noserverrestore": False,
    },
}


class BUIServer(Flask):
    """
    The :class:`burpui.engines.server.BUIServer` class provides the ``Burp-UI`` server.
    """

    gunicorn = False

    def __init__(self):
        """The :class:`burpui.engines.server.BUIServer` class provides the ``Burp-UI``
        server.

        :param app: The Flask application to launch
        """
        super(BUIServer, self).__init__("burpui")
        self.init = False
        # We cannot override the Flask's logger so we use our own
        self._logger = logger
        self._logger.disabled = True
        self.conf = config
        # switch the flask config with our magic config object
        self.conf.update(self.config)
        self.config = self.conf

    def enable_logger(self, enable=True):
        """Enable or disable the logger"""
        self._logger.disabled = not enable

    @property
    def logger(self):
        """
        :rtype: :class:`logging.Logger`
        """
        return self._logger

    def setup(self, conf=None, unittest=False, cli=False):
        """The :func:`burpui.engines.server.BUIServer.setup` functions is used to setup
        the whole server by parsing the configuration file and loading the
        different backends.

        :param conf: Path to a configuration file
        :type conf: str

        :param unittest: Whether we are unittesting or not (used to avoid burp2
                         strict requirements checks)
        :type unittest: bool

        :param cli: Whether we are in cli mode or not
        :type cli: bool
        """
        self.sslcontext = None
        if not conf:
            conf = self.config["CFG"]

        if not conf:
            raise IOError("No configuration file found")

        # Raise exception if errors are encountered during parsing
        self.conf.parse(conf, BUI_DEFAULTS)
        self.conf.default_section("Global")

        self.config["BUI_BIND"] = self.conf.safe_get("bind")
        self.config["BUI_PORT"] = self.conf.safe_get("port", "integer")
        if self.config["BUI_BIND"] or self.config["BUI_PORT"]:
            warnings.warn(
                "The 'bind' and 'port' configuration options are now deprecated and "
                "have no effect on burp-ui anymore unless you use the 'burp-ui-legacy' "
                "command.\n"
                "Please use the '-h' and '-p' command line flags instead.",
                UserWarning,
            )
        self.demo = self.config["BUI_DEMO"] = self.conf.safe_get("demo", "boolean")
        self.config["BUI_DSN"] = self.conf.safe_get("dsn")
        self.config["BUI_PIWIK_URL"] = self.conf.safe_get("piwik_url")
        self.config["BUI_PIWIK_SCRIPT"] = self.conf.safe_get("piwik_script")
        self.config["BUI_PIWIK_ID"] = self.conf.safe_get("piwik_id", "integer")
        self.config["BACKEND"] = self.conf.safe_get("backend")
        # FIXME: this sucks, we need to test the burp2 backend as well!
        if unittest:
            self.config["BACKEND"] = "burp1"
        self.config["STANDALONE"] = self.config["BACKEND"] != "multi"
        self.config["BUI_SSL"] = self.conf.safe_get("ssl", "boolean")
        self.config["BUI_SSLCERT"] = self.conf.safe_get("sslcert")
        self.config["BUI_SSLKEY"] = self.conf.safe_get("sslkey")
        if (
            self.config["BUI_SSL"]
            or self.config["BUI_SSLCERT"]
            or self.config["BUI_SSLKEY"]
        ):
            warnings.warn(
                "The 'ssl', 'sslcert' and 'sslkey' configuration options are deprecated "
                "as of v0.4.0. It means they have no effect on burp-ui anymore. "
                "If you really need them, consider using the 'burp-ui-legacy' command "
                "instead.",
                UserWarning,
            )

        # TODO: remove in 0.8.0
        old_prefix = self.conf.safe_get("prefix")

        self.plugins = self.config["BUI_PLUGINS"] = self.conf.safe_get(
            "plugins", "string_lower_list"
        )
        if len(self.plugins) == 1 and self.plugins[0] == "none":
            self.plugins = self.config["BUI_PLUGINS"] = []

        self.auth = self.config["BUI_AUTH"] = self.conf.safe_get(
            "auth", "string_lower_list"
        )

        self.audit_backends = self.config["BUI_AUDIT"] = self.conf.safe_get(
            "audit", "string_lower_list"
        )

        # UI options
        self.config["REFRESH"] = self.conf.safe_get("refresh", "integer", "UI")
        self.config["LIVEREFRESH"] = self.conf.safe_get("liverefresh", "integer", "UI")
        self.config["DEFAULT_STRIP"] = self.conf.safe_get(
            "default_strip", "integer", "UI"
        )
        self.ignore_labels = self.conf.safe_get("ignore_labels", "force_list", "UI")
        format_labels = self.conf.safe_get("format_labels", "force_list", "UI")
        self.format_labels = []
        for format_label in format_labels:
            search = re.search(
                r"^s(?P<separator>.)(?P<regex>.*?)(?P=separator)(?P<replace>.*?)(?P=separator)$",
                format_label,
            )
            if search:
                self.format_labels.append(
                    (search.group("regex"), search.group("replace"))
                )

        # Production options
        self.storage = self.config["BUI_STORAGE"] = self.conf.safe_get(
            "storage", section="Production"
        )
        self.cache_db = self.config["BUI_CACHE_DB"] = self.conf.safe_get(
            "cache", section="Production"
        )
        self.session_db = self.config["BUI_SESSION_DB"] = self.conf.safe_get(
            "session", section="Production"
        )
        self.redis = self.config["BUI_REDIS"] = self.conf.safe_get(
            "redis", section="Production"
        )
        self.limiter = self.config["BUI_LIMITER"] = self.conf.safe_get(
            "limiter", "boolean_or_string", section="Production"
        )
        if isinstance(self.limiter, bool) and not self.limiter:
            self.limiter = self.config["BUI_LIMITER"] = "none"
        self.config["BUI_RATIO"] = self.conf.safe_get("ratio", section="Production")
        self.use_celery = self.config["BUI_CELERY"] = self.conf.safe_get(
            "celery", "boolean_or_string", section="Production"
        )
        self.database = self.config["SQLALCHEMY_DATABASE_URI"] = self.conf.safe_get(
            "database", "boolean_or_string", section="Production"
        )
        self.config["WITH_LIMIT"] = False
        if isinstance(self.database, bool):
            self.config["WITH_SQL"] = self.database
        else:
            self.config["WITH_SQL"] = self.database and self.database.lower() != "none"
        if isinstance(self.use_celery, bool):
            self.config["WITH_CELERY"] = self.use_celery
        else:
            self.config["WITH_CELERY"] = (
                self.use_celery and self.use_celery.lower() != "none"
            )
        self.config["NUM_PROXIES"] = self.conf.safe_get(
            "num_proxies", "integer", section="Production"
        )
        self.config["PROXY_FIX_ARGS"] = self.conf.safe_get(
            "proxy_fix_args", section="Production"
        )
        prefix = self.conf.safe_get("prefix", section="Production")
        if not prefix and old_prefix:
            # TODO: remove in a later version
            prefix = old_prefix
            warnings.warn(
                "The 'prefix' option has been moved from the '[Global]' section to the "
                "'[Production]' section",
                UserWarning,
            )
        if prefix and not prefix.startswith("/"):
            if prefix.lower() != "none":
                self.logger.warning("'prefix' must start with a '/'!")
            prefix = ""
        self.config["BUI_PREFIX"] = prefix

        # WebSocket options
        self.ws_enabled = self.config["WS_ENABLED"] = self.conf.safe_get(
            "enabled", "boolean", section="WebSocket"
        )
        self.config["WITH_WS"] = self.conf.safe_get(
            "embedded", "boolean", section="WebSocket"
        )
        self.ws_broker = self.config["BUI_WS_BROKER"] = self.conf.safe_get(
            "broker", "boolean_or_string", section="WebSocket"
        )
        self.config["WS_DEBUG"] = self.conf.safe_get(
            "debug", "boolean", section="WebSocket"
        )
        self.config["WS_URL"] = self.conf.safe_get("url", section="WebSocket")
        if self.config.get("WS_URL", "").lower() == "none" or self.config.get(
            "WITH_WS", False
        ):
            self.config["WS_URL"] = None

        # Experimental options
        self.noserverrestore = self.conf.safe_get(
            "noserverrestore", "boolean", section="Experimental"
        )

        # Security options
        self.config["BUI_SCOOKIE"] = self.conf.safe_get(
            "scookie", "boolean", section="Security"
        )
        self.config["SECRET_KEY"] = self.conf.safe_get("appsecret", section="Security")
        days = self.conf.safe_get("cookietime", "integer", section="Security") or 14
        self.config["REMEMBER_COOKIE_DURATION"] = self.config[
            "PERMANENT_SESSION_LIFETIME"
        ] = timedelta(days=days)
        self.config["REMEMBER_COOKIE_NAME"] = "remember_token"
        days = self.conf.safe_get("sessiontime", "integer", section="Security") or 5
        self.config["SESSION_INACTIVE"] = timedelta(days=days)

        self.logger.info("backend: {}".format(self.config["BACKEND"]))
        self.logger.info("listen port: {}".format(self.config["BUI_PORT"]))
        self.logger.info("bind addr: {}".format(self.config["BUI_BIND"]))
        self.logger.info("use ssl: {}".format(self.config["BUI_SSL"]))
        self.logger.info("standalone: {}".format(self.config["STANDALONE"]))
        self.logger.info("sslcert: {}".format(self.config["BUI_SSLCERT"]))
        self.logger.info("sslkey: {}".format(self.config["BUI_SSLKEY"]))
        self.logger.info("prefix: {}".format(self.config["BUI_PREFIX"]))
        self.logger.info("secure cookie: {}".format(self.config["BUI_SCOOKIE"]))
        self.logger.info(
            "cookietime: {}".format(self.config["REMEMBER_COOKIE_DURATION"])
        )
        self.logger.info("session inactive: {}".format(self.config["SESSION_INACTIVE"]))
        self.logger.info("refresh: {}".format(self.config["REFRESH"]))
        self.logger.info("liverefresh: {}".format(self.config["LIVEREFRESH"]))
        self.logger.info("auth: {}".format(self.auth))
        self.logger.info("audit: {}".format(self.audit_backends))
        self.logger.info("celery: {}".format(self.use_celery))
        self.logger.info("redis: {}".format(self.redis))
        self.logger.info("limiter: {}".format(self.limiter))
        self.logger.info("database: {}".format(self.database))
        self.logger.info("with SQL: {}".format(self.config["WITH_SQL"]))
        self.logger.info("with Celery: {}".format(self.config["WITH_CELERY"]))
        self.logger.info("with WebSocket: {}".format(self.config["WITH_WS"]))
        self.logger.info("demo: {}".format(self.config["BUI_DEMO"]))

        self.init = True
        if not cli:
            self.load_modules()

    def load_modules(self, strict=False):
        """Load the extensions"""
        self.plugin_manager = PluginManager(self, self.plugins)
        if self.plugins:
            self.plugin_manager.load_all()

        try:
            self.audit = BUIauditLoader(self)
        except ImportError as exc:
            self.logger.critical(
                f"Import Exception, module '{self.audit_backends}': {exc}"
            )
            raise exc

        if self.auth and "none" not in self.auth:
            try:
                self.uhandler = UserAuthHandler(self)
                for back, err in self.uhandler.errors.items():
                    self.logger.critical(
                        "Unable to load '{}' authentication backend:\n{}".format(
                            back, err
                        )
                    )
            except ImportError as exc:
                self.logger.critical(
                    "Import Exception, module '{0}': {1}".format(self.auth, str(exc))
                )
                raise exc
            self.acl_engine = self.config["BUI_ACL"] = self.conf.safe_get(
                "acl", "string_lower_list"
            )
            self.config["LOGIN_DISABLED"] = False
        else:
            self.config["LOGIN_DISABLED"] = True
            # No login => no ACL
            self.acl_engine = self.config["BUI_ACL"] = ["none"]
            self.auth = self.config["BUI_AUTH"] = "none"

        if self.acl_engine and "none" not in self.acl_engine:
            try:
                self.acl_handler = ACLloader(self)
            except Exception as exc:
                self.logger.critical(
                    "Import Exception, module '{0}': {1}".format(
                        self.acl_engine, str(exc)
                    )
                )
                raise exc
        else:
            self.acl_handler = False

        self.logger.info("acl: {}".format(self.acl_engine))

        backend = self.config["BACKEND"]
        if "." in backend:
            module = backend
        else:
            module = "burpui.misc.backend.{}".format(backend)

        # This is used for development purpose only
        from ..misc.backend.burp1 import Burp as BurpGeneric

        self.client = BurpGeneric(dummy=True)
        self.strict = strict
        try:
            # lookup plugins first
            mod = self.plugin_manager.get_plugin_by_name(backend)
            if mod:
                self.client = mod.Burp(self, conf=self.conf)
            else:
                # Try to load submodules from our current environment
                # first
                sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
                mod = __import__(module, fromlist=["Burp"])
                self.client = mod.Burp(self, conf=self.conf)
        except Exception as exc:
            msg = "Failed loading backend {0}: {1}".format(module, str(exc))
            if strict:
                self.logger.critical(msg, exc_info=exc, stack_info=True)
                sys.exit(2)
            else:
                exc.args = (msg,)
                raise exc

        self.audit.logger.info("Burp-UI server started")

    @property
    def acl(self):
        """ACL module

        :returns: :class:`burpui.misc.acl.interface.BUIacl`
        """
        if self.acl_engine and "none" not in self.acl_engine:
            # refresh acl to detect config changes
            from ..misc.acl.interface import BUIacl  # noqa

            acl = self.acl_handler.acl  # type: BUIacl
            return acl
        return None

    def get_send_file_max_age(self, name):
        """Provides default cache_timeout for the send_file() functions."""
        if name:
            lname = name.lower()
            extensions = ["js", "css", "woff"]
            for ext in extensions:
                if lname.endswith(".{}".format(ext)):
                    return 3600 * 24 * 30  # 30 days
        return Flask.get_send_file_max_age(self, name)

    def manual_run(self):
        """The :func:`burpui.engines.server.BUIServer.manual_run` functions is used to
        actually launch the ``Burp-UI`` server.

        .. deprecated:: 0.4.0
           You should now run the Flask's run command.
        """
        if not self.init:
            self.setup()

        if self.config["BUI_SSL"]:
            self.sslcontext = (self.config["BUI_SSLCERT"], self.config["BUI_SSLKEY"])

        if self.sslcontext:
            self.config["SSL"] = True
            self.run(
                host=self.config["BUI_BIND"],
                port=self.config["BUI_PORT"],
                debug=self.config["DEBUG"],
                ssl_context=self.sslcontext,
            )
        else:
            self.run(
                host=self.config["BUI_BIND"],
                port=self.config["BUI_PORT"],
                debug=self.config["DEBUG"],
            )



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom.fields
    :platform: Unix
    :synopsis: Burp-UI api custom fields module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>


"""
import flask_restx.fields

from flask_restx.fields import *  # noqa # pylint: disable=locally-disabled, wildcard-import, unused-wildcard-import
from .my_fields import (
    DateTime,
    DateTimeHuman,
    BackupNumber,
    SafeString,
    LocalizedString,
)  # noqa

__all__ = flask_restx.fields.__all__ + (
    DateTime,
    DateTimeHuman,
    BackupNumber,
    SafeString,
    LocalizedString,
)



# -*- coding: utf8 -*-
from flask_login import AnonymousUserMixin

from .interface import BUIhandler, BUIuser, BUIloader
from ...utils import __

import ssl

try:
    from ldap3 import (
        Server,
        Connection,
        Tls,
        ALL,
        RESTARTABLE,
        AUTO_BIND_TLS_BEFORE_BIND,
        AUTO_BIND_NONE,
        SIMPLE,
    )
except ImportError:
    raise ImportError("Unable to load 'ldap3' module")


class LdapLoader(BUIloader):
    """The :class:`burpui.misc.auth.ldap.LdapLoader` handles searching for and
    binding as a :class:`burpui.misc.auth.ldap.LdapUser` user.
    """

    section = name = "LDAP:AUTH"

    def __init__(self, app=None, handler=None):
        """:func:`burpui.misc.auth.ldap.LdapLoader.__init__` establishes a
        connection to the LDAP server.

        :param app: Instance of the app we are running in
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = app
        conf = self.app.conf
        handler.name = self.name
        defaults = {
            "LDAP:AUTH": {
                "host": "localhost",
                "port": None,
                "encryption": None,
                "binddn": None,
                "bindpw": None,
                "filter": None,
                "base": None,
                "searchattr": "uid",
                "validate": "none",
                "cafile": None,
            }
        }
        mapping = {
            "host": "host",
            "port": "port",
            "encryption": "encryption",
            "filt": "filter",
            "base": "base",
            "attr": "searchattr",
            "binddn": "binddn",
            "bindpw": "bindpw",
            "validate": "validate",
            "cafile": "cafile",
        }
        conf.update_defaults(defaults)
        # Maybe the handler argument is None, maybe the 'priority'
        # option is missing. We don't care.
        try:
            handler.priority = (
                conf.safe_get("priority", "integer", section=self.section)
                or handler.priority
            )
        except:
            pass
        for (opt, key) in mapping.items():
            setattr(self, opt, conf.safe_get(key, "force_string", section=self.section))

        if self.validate and self.validate.lower() in ["none", "optional", "required"]:
            self.validate = getattr(ssl, "CERT_{}".format(self.validate.upper()))
        else:
            self.validate = None
        self.version = ssl.OP_NO_SSLv3
        self.users = []
        self.tls = None
        self.ssl = False
        self.auto_bind = AUTO_BIND_NONE
        if self.encryption == "ssl":
            self.ssl = True
        elif self.encryption == "tls":
            self.tls = Tls(
                local_certificate_file=self.cafile,
                validate=self.validate,
                version=self.version,
            )
            self.auto_bind = AUTO_BIND_TLS_BEFORE_BIND
        if self.port:
            try:
                self.port = int(self.port)
            except ValueError:
                self.logger.error("LDAP port must be a valid integer")
                self.port = None
        self.logger.info("LDAP host: {0}".format(self.host))
        self.logger.info("LDAP port: {0}".format(self.port))
        self.logger.info("LDAP encryption: {0}".format(self.encryption))
        self.logger.info("LDAP filter: {0}".format(self.filt))
        self.logger.info("LDAP base: {0}".format(self.base))
        self.logger.info("LDAP search attr: {0}".format(self.attr))
        self.logger.info("LDAP binddn: {0}".format(self.binddn))
        self.logger.info("LDAP bindpw: {0}".format("*****" if self.bindpw else "None"))
        self.logger.info("TLS object: {0}".format(self.tls))

        try:
            self.server = Server(
                host=self.host,
                port=self.port,
                use_ssl=self.ssl,
                get_info=ALL,
                tls=self.tls,
            )
            self.logger.debug("LDAP Server = {0}".format(str(self.server)))
            if self.binddn:
                self.ldap = Connection(
                    self.server,
                    user=self.binddn,
                    password=self.bindpw,
                    raise_exceptions=True,
                    client_strategy=RESTARTABLE,
                    auto_bind=self.auto_bind,
                    authentication=SIMPLE,
                )
            else:
                self.ldap = Connection(
                    self.server,
                    raise_exceptions=True,
                    client_strategy=RESTARTABLE,
                    auto_bind=self.auto_bind,
                )
            okay = False
            with self.ldap:
                self.logger.debug("LDAP Connection = {0}".format(str(self.ldap)))
                self.logger.info("OK, connected to LDAP")
                okay = True

            if not okay:
                raise Exception("Not connected")

            self._prefetch()
        except Exception as e:
            self.logger.error("Could not connect to LDAP: {0}".format(str(e)))
            self.server = None
            self.ldap = None

    def __exit__(self, exc_type, exc_value, traceback):
        """:func:`burpui.misc.auth.ldap.LdapLoader.__exit__` closes the
        connection to the LDAP server.
        """
        if self.ldap and self.ldap.bound:
            self.ldap.unbind()

    def fetch(self, searchval=None, uniq=True):
        """:func:`burpui.misc.auth.ldap.LdapLoader.fetch` searches for a user
        object in the LDAP server.

        :param searchval: attribute value to search for
        :type searchval: str

        :param uniq: only return one result
        :type uniq: bool

        :returns: dictionary of `distinguishedName` and `commonName` attributes for the
        user if found, otherwise None.
        """
        try:
            if self.filt:
                query = self.filt.format(self.attr, searchval)
            else:
                query = "({0}={1})".format(self.attr, searchval)
            self.logger.info("filter: {0} | base: {1}".format(query, self.base))
            r = None
            with self.ldap:
                self.logger.debug("LDAP Connection = {0}".format(str(self.ldap)))
                self.ldap.search(self.base, query, attributes=["cn", self.attr])
                r = self.ldap.response
            if not r:
                raise ValueError("no results")
        except Exception as e:
            self.logger.error("Ooops, LDAP lookup failed: {0}".format(str(e)))
            return None

        if not uniq:
            return r

        for record in r:
            attrs = record["attributes"]
            if self.attr in attrs and searchval in attrs[self.attr]:
                self.logger.info("Found DN: {0}".format(record["dn"]))
                return {"dn": record["dn"], "cn": attrs["cn"][0]}

    def _prefetch(self):
        """Prefetch all users that match the filter/base"""
        self.users = []
        results = self.fetch("*", False) or []
        for record in results:
            attrs = record["attributes"]
            if self.attr in attrs:
                self.users.append(attrs[self.attr][0])
        self.logger.debug(self.users)

    def check(self, dn=None, passwd=None):
        """:func:`burpui.misc.auth.ldap.LdapLoader.check` authenticates a user
        against the LDAP server.

        :param dn: canonical `dn` of the user to authenticate as
        :type dn: str

        :param passwd: password of the user to authenticate as
        :type passwd: str

        :returns: True if bind was successful, otherwise False
        """
        try:
            with Connection(
                self.server,
                user="{0}".format(dn),
                password=passwd,
                raise_exceptions=True,
                auto_bind=self.auto_bind,
                authentication=SIMPLE,
            ) as con:
                self.logger.debug("LDAP Connection = {0}".format(str(con)))
                self.logger.info("Bound as user: {0}".format(dn))
                return con.bind()
        except Exception as e:
            self.logger.error(
                "Failed to authenticate user: {0}, {1}".format(dn, str(e))
            )

        self.logger.error("Bind as '{0}' failed".format(dn))
        return False


class UserHandler(BUIhandler):
    __doc__ = __(
        "Connects to a LDAP database to authenticate users. Handles "
        "searching for and binding as."
    )
    priority = 50

    preload_users = False

    """The :class:`burpui.misc.auth.ldap.UserHandler` class maintains a list of
    ``Burp-UI`` users.
    """

    def __init__(self, app=None):
        """:func:`burpui.misc.auth.ldap.UserHandler.__init__` creates the
        handler instance

        :param app: Instance of the app we are running in
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.ldap = LdapLoader(app, self)
        self.users = {}

    def user(self, name=None):
        """See :func:`burpui.misc.auth.interface.BUIhandler.user`"""
        if name not in self.users:
            self.users[name] = LdapUser(self.ldap, name)
        ret = self.users[name]
        if not ret.active:
            return AnonymousUserMixin()
        return ret

    @property
    def loader(self):
        return self.ldap


class LdapUser(BUIuser):
    """The :class:`burpui.misc.auth.ldap.LdapUser` class generates a ``Burp-UI``
    user from a user object found in the LDAP server.
    """

    def __init__(self, ldap=None, name=None):
        """:func:`burpui.misc.auth.ldap.LdapUser.__init__` function finds a user
        in the LDAP server and stores the DN of the user if found.

        :param ldap: an ``LdapLoader`` instance
        :type ldap: :class:`burpui.misc.auth.ldap.LdapLoader`

        :param name: login name of the user to find in the LDAP server
        :param type: str
        """
        self.active = False
        self.authenticated = False
        self.ldap = ldap
        self.name = name
        self.backend = self.ldap.name

        found = self.ldap.fetch(name)

        if found:
            self.id = found["dn"]
            self.active = True

    def login(self, passwd=None):
        """:func:`burpui.misc.auth.ldap.LdapUser.login` function finds a user in
        the LDAP server and authenticates that user using an LDAP bind.

        :param passwd: password to bind to the LDAP server with
        :type passwd: str

        :returns: True if found and bind was successful;
                  False if found but bind failed;
                  otherwise de-activates the user and returns False
        """
        if self.ldap.fetch(self.name):
            self.authenticated = self.ldap.check(self.id, passwd)
            return self.authenticated
        else:
            self.authenticated = False
            self.active = False
            return False

    def get_id(self):
        """:func:`burpui.misc.auth.ldap.LdapUser.get_id` function

        :returns: login name of the user
        """
        return self.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

from subprocess import check_output, call, STDOUT
from distutils import log
from distutils.core import Command
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.sdist import sdist
from setuptools.command.install import install
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.egg_info import egg_info

ROOT = os.path.join(os.path.dirname(os.path.realpath(__file__)))
DEVNULL = open(os.devnull, "wb")

# Not sure bower was a great idea...
VENDOR_TO_KEEP = [
    "burpui/static/vendor/bootswatch/slate/bootstrap.min.css",
    "burpui/static/vendor/bootswatch/fonts/glyphicons-halflings-regular.eot",
    "burpui/static/vendor/bootswatch/fonts/glyphicons-halflings-regular.svg",
    "burpui/static/vendor/bootswatch/fonts/glyphicons-halflings-regular.ttf",
    "burpui/static/vendor/bootswatch/fonts/glyphicons-halflings-regular.woff",
    "burpui/static/vendor/bootswatch/fonts/glyphicons-halflings-regular.woff2",
    "burpui/static/vendor/nvd3/build/nv.d3.min.css",
    "burpui/static/vendor/datatables.net-bs/css/dataTables.bootstrap.min.css",
    "burpui/static/vendor/datatables.net-responsive-bs/css/responsive.bootstrap.min.css",
    "burpui/static/vendor/datatables.net-select-bs/css/select.bootstrap.min.css",
    "burpui/static/vendor/datatables.net-buttons-bs/css/buttons.bootstrap.min.css",
    "burpui/static/vendor/datatables.net-fixedheader-bs/css/fixedHeader.bootstrap.min.css",
    "burpui/static/vendor/jquery.fancytree/dist/skin-bootstrap/ui.fancytree.min.css",
    "burpui/static/vendor/bootstrap-switch/dist/css/bootstrap3/bootstrap-switch.min.css",
    "burpui/static/vendor/angular-ui-select/dist/select.min.css",
    "burpui/static/vendor/jquery/dist/jquery.min.js",
    "burpui/static/vendor/jquery-ui/jquery-ui.min.js",
    "burpui/static/vendor/bootstrap/dist/js/bootstrap.min.js",
    "burpui/static/vendor/typeahead.js/dist/typeahead.jquery.min.js",
    "burpui/static/vendor/d3/d3.min.js",
    "burpui/static/vendor/nvd3/build/nv.d3.min.js",
    "burpui/static/vendor/datatables.net/js/jquery.dataTables.min.js",
    "burpui/static/vendor/datatables.net-bs/js/dataTables.bootstrap.min.js",
    "burpui/static/vendor/datatables.net-responsive/js/dataTables.responsive.min.js",
    "burpui/static/vendor/datatables.net-responsive-bs/js/responsive.bootstrap.min.js",
    "burpui/static/vendor/datatables.net-select/js/dataTables.select.min.js",
    "burpui/static/vendor/datatables.net-buttons/js/dataTables.buttons.min.js",
    "burpui/static/vendor/datatables.net-buttons-bs/js/buttons.bootstrap.min.js",
    "burpui/static/vendor/datatables.net-fixedheader/js/dataTables.fixedHeader.min.js",
    "burpui/static/vendor/jquery.floatThead/dist/jquery.floatThead.min.js",
    "burpui/static/vendor/jquery.fancytree/dist/jquery.fancytree-all.min.js",
    "burpui/static/vendor/jquery-file-download/src/Scripts/jquery.fileDownload.js",
    "burpui/static/vendor/lodash/dist/lodash.min.js",
    "burpui/static/vendor/angular/angular.min.js",
    "burpui/static/vendor/angular-route/angular-route.min.js",
    "burpui/static/vendor/angular-sanitize/angular-sanitize.min.js",
    "burpui/static/vendor/angular-resource/angular-resource.min.js",
    "burpui/static/vendor/angular-animate/angular-animate.min.js",
    "burpui/static/vendor/bootstrap-switch/dist/js/bootstrap-switch.min.js",
    "burpui/static/vendor/angular-bootstrap-switch/dist/angular-bootstrap-switch.min.js",
    "burpui/static/vendor/angular-ui-select/dist/select.min.js",
    "burpui/static/vendor/angular-strap/dist/angular-strap.min.js",
    "burpui/static/vendor/angular-strap/dist/angular-strap.tpl.min.js",
    "burpui/static/vendor/angular-onbeforeunload/build/angular-onbeforeunload.js",
    "burpui/static/vendor/angular-datatables-0.6.2/dist/angular-datatables.min.js",
    "burpui/static/vendor/angular-highlightjs/build/angular-highlightjs.min.js",
    "burpui/static/vendor/moment/min/moment.min.js",
    "burpui/static/vendor/moment/locale/fr.js",
    "burpui/static/vendor/moment/locale/es.js",
    "burpui/static/vendor/moment/locale/it.js",
    "burpui/static/vendor/moment-timezone/builds/moment-timezone-with-data-10-year-range.min.js",
    "burpui/static/vendor/angular-ui-calendar/src/calendar.js",
    "burpui/static/vendor/fullcalendar/dist/fullcalendar.min.css",
    "burpui/static/vendor/fullcalendar/dist/fullcalendar.print.min.css",
    "burpui/static/vendor/fullcalendar/dist/fullcalendar.min.js",
    "burpui/static/vendor/fullcalendar/dist/gcal.min.js",
    "burpui/static/vendor/fullcalendar/dist/locale/fr.js",
    "burpui/static/vendor/fullcalendar/dist/locale/es.js",
    "burpui/static/vendor/fullcalendar/dist/locale/it.js",
    "burpui/static/vendor/angular-bootstrap/ui-bootstrap.min.js",
    "burpui/static/vendor/angular-bootstrap/ui-bootstrap-tpls.min.js",
    "burpui/static/vendor/components-font-awesome/css/font-awesome.min.css",
    "burpui/static/vendor/components-font-awesome/fonts/FontAwesome.otf",
    "burpui/static/vendor/components-font-awesome/fonts/fontawesome-webfont.eot",
    "burpui/static/vendor/components-font-awesome/fonts/fontawesome-webfont.svg",
    "burpui/static/vendor/components-font-awesome/fonts/fontawesome-webfont.ttf",
    "burpui/static/vendor/components-font-awesome/fonts/fontawesome-webfont.woff",
    "burpui/static/vendor/components-font-awesome/fonts/fontawesome-webfont.woff2",
    "burpui/static/vendor/socket.io-client/dist/socket.io.min.js",
    "burpui/static/vendor/js-cookie/src/js.cookie.js",
    "burpui/static/vendor/ace-builds/src-min-noconflict/ace.js",
    "burpui/static/vendor/ace-builds/src-min-noconflict/mode-json.js",
    "burpui/static/vendor/ace-builds/src-min-noconflict/worker-json.js",
    "burpui/static/vendor/ace-builds/src-min-noconflict/theme-ambiance.js",
]

for p in VENDOR_TO_KEEP:
    if not os.path.exists(p):
        log.info("!! missing: {}".format(p))


class DevelopWithBuildStatic(develop):
    def install_for_development(self):
        self.run_command("build_static")
        return develop.install_for_development(self)


class EggWithBuildStatic(egg_info):
    def initialize_options(self):
        self.run_command("build_static")
        return egg_info.initialize_options(self)


class BdistWithBuildStatic(bdist_egg):
    def initialize_options(self):
        self.run_command("build_static")
        return bdist_egg.initialize_options(self)


class SdistWithBuildStatic(sdist):
    def make_distribution(self):
        self.run_command("build_static")
        return sdist.make_distribution(self)


class PyTest(Command):
    user_options = []
    description = "Run tests"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            errno = call(["make", "test"])
            raise SystemExit(errno)
        except OSError:
            log.error("Looks like the tools to run the tests are missing")


class BuildStatic(Command):
    user_options = []
    description = "Install bower dependencies"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.chdir(ROOT)
        log.info("compiling translations")
        call(
            "{} ./burpui -m manage compile-translation".format(sys.executable).split(),
            stderr=DEVNULL,
        )
        log.info("getting revision number")
        rev = "stable"
        ci = os.getenv("CI")
        commit = os.getenv("CI_COMMIT_SHA")
        if (
            not ci
            and os.path.exists(".git")
            and call("which git", shell=True, stderr=STDOUT, stdout=DEVNULL) == 0
        ):
            try:
                branch = check_output("git rev-parse HEAD", shell=True).rstrip()
                ver = open(os.path.join("burpui", "VERSION")).read().rstrip()
                log.info("version: {}".format(ver))
                if branch and "dev" in ver:
                    rev = branch
                try:
                    log.info("revision: {}".format(rev))
                    with open("burpui/RELEASE", "wb") as f:
                        f.write(rev)
                except:
                    log.error("Unable to create release file")
            except:
                pass
        elif ci:
            try:
                ver = open(os.path.join("burpui", "VERSION")).read().rstrip()
                if "dev" in ver:
                    rev = commit
                try:
                    with open("burpui/RELEASE", "wb") as f:
                        f.write(rev)
                except:
                    pass
            except:
                pass
        else:
            log.info("using upstream revision")
        keep = VENDOR_TO_KEEP
        dirlist = []
        for dirname, subdirs, files in os.walk("burpui/static/vendor"):
            for filename in files:
                path = os.path.join(dirname, filename)
                _, ext = os.path.splitext(path)
                if (
                    os.path.isfile(path)
                    and path not in keep
                    and filename not in ["bower.json", "package.json"]
                ):
                    if (rev == "stable" and ext == ".map") or ext != ".map":
                        os.unlink(path)
                elif os.path.isdir(path):
                    dirlist.append(path)
        dirlist.sort(reverse=True)
        for d in dirlist:
            if os.path.isdir(d) and not os.listdir(d):
                os.rmdir(d)


class CustomInstall(install):
    def run(self):
        self.run_command("build_static")
        install.run(self)


def readme():
    """
    Function used to skip the screenshots part
    """
    desc = ""
    cpt = 0
    skip = False
    with open(os.path.join(ROOT, "README.rst")) as f:
        for line in f.readlines():
            if line.rstrip() == "Screenshots":
                skip = True
            if skip:
                cpt += 1
            if cpt > 6:
                skip = False
            if skip:
                continue
            desc += line
    return desc


sys.path.insert(0, os.path.join(ROOT))

from burpui.desc import (
    __author__,
    __author_email__,
    __description__,
    __url__,
    __title__,
)  # noqa

name = __title__
author = __author__
author_email = __author_email__
description = __description__
url = __url__

with open(os.path.join(ROOT, "requirements.txt")) as f:
    requires = [x.strip() for x in f if x.strip()]

dev_requires = ["flake8", "pylint", "black"]
test_requires = [
    "pytest",
    "pytest-cov",
    "pytest-flask",
    "pytest-mock",
    "nose",
    "coverage",
    "mock",
    "mockredispy",
    "Flask-Session",
    "Celery",
    "redis",
    "Flask-SQLAlchemy",
    "Flask-Migrate",
    "sqlalchemy_utils",
]

datadir = os.path.join("share", "burpui")
confdir = os.path.join(datadir, "etc")
contrib = os.path.join(datadir, "contrib")
migrations = [
    (
        os.path.join(datadir, root),
        [os.path.join(root, f) for f in files if not f.endswith(".pyc")],
    )
    for root, dirs, files in os.walk("migrations")
]

setup(
    name=name,
    version=open(os.path.join(ROOT, "burpui", "VERSION")).read().rstrip(),
    description=description,
    long_description=readme(),
    license=open(os.path.join(ROOT, "LICENSE")).readline().rstrip(),
    author=author,
    author_email=author_email,
    url=url,
    keywords="burp web ui backup monitoring",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "static": ["burpui/static/*"],
        "templates": ["burpui/templates/*"],
        "VERSION": ["burpui/VERSION"],
    },
    entry_points={
        "console_scripts": [
            "burp-ui=burpui.__main__:server",
            "bui-celery=burpui.__main__:celery",
            "bui-manage=burpui.__main__:manage",
            "bui-agent-legacy=burpui.__main__:agent",
            "bui-monitor=burpui.__main__:monitor",
            "burp-ui-legacy=burpui.__main__:legacy",
        ],
    },
    data_files=[
        (confdir, [os.path.join(confdir, "burpui.sample.cfg")]),
        (os.path.join(contrib, "centos"), ["contrib/centos/init.sh"]),
        (os.path.join(contrib, "freebsd"), ["contrib/freebsd/gunicorn.rc"]),
        (os.path.join(contrib, "gunicorn"), ["contrib/gunicorn/burpui_gunicorn.py"]),
        (
            os.path.join(contrib, "systemd"),
            [
                "contrib/systemd/bui-agent.service",
                "contrib/systemd/bui-celery.service",
                "contrib/systemd/bui-celery-beat.service",
                "contrib/systemd/bui-gunicorn.service",
                "contrib/systemd/bui-monitor.service",
                "contrib/systemd/bui-websocket.service",
            ],
        ),
    ]
    + migrations,
    python_requires=">=3.7",
    install_requires=requires,
    extras_require={
        "ldap_authentication": ["ldap3"],
        "gunicorn": ["gunicorn"],
        "gunicorn-extra": ["redis"],
        "ci": test_requires,
        "dev": dev_requires,
        "celery": ["Celery>=4.3", "redis"],
        "sql": ["Flask-SQLAlchemy", "Flask-Migrate>=2.1.0", "sqlalchemy-utils"],
        "limit": ["Flask-Limiter", "redis"],
        "websocket": ["flask-socketio", "redis", "gevent-websocket"],
        "rtd": ["sphinxcontrib-httpdomain==1.8.0"],
    },
    tests_require=test_requires,
    classifiers=[
        "Framework :: Flask",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: System :: Monitoring",
    ],
    cmdclass={
        "build_static": BuildStatic,
        "develop": DevelopWithBuildStatic,
        "sdist": SdistWithBuildStatic,
        "install": CustomInstall,
        "bdist_egg": BdistWithBuildStatic,
        "egg_info": EggWithBuildStatic,
    },
)
